#ifndef _MD_RFUNCTION_C_
#define _MD_RFUNCTION_C_
#include "util.h"
#include "config.h"
/**
 *  this function an array with the value for the
 *   integral from 0 to 1 of  pow(u,2*m)exp(-t*u*u) du.
 */
inline static void
fgamma (double *fm, double t, int m)
{
  register int i;
  register double sum, term, expt, twot, mphalf;
  const double tcrit = 36. + 2 * m;
  const double sqrtpi = 0.88622692545275801365;
  const double eps = 1.e-14;
  if (t > tcrit)
    {
      fm[0] = sqrtpi / sqrt (t);
      for (i = 1; i <= m; i++)
	fm[i] = fm[i - 1] * (i - 0.5) / t;
      return;
    }
  mphalf = m + 0.5;
  term = 0.5 / mphalf;
  sum = term;
  for (i = 1; i <= 500; i++)
    {
      term *= (t / (mphalf + i));
      sum += term;
      if (term < eps)
	break;
    }
  twot = 2.0 * t;
  expt = exp (-t);
  fm[m] = sum * expt;
  for (i = m; i--;)
    fm[i] = (fm[i + 1] * twot + expt) / (2.0 * i + 1.0);
}


void
formMDR (double ***r, double sr, double t, double w,
	 const double *pq, int ltot)
{

  register int m;
  register double term, sterm, x, y, z, x2, y2, z2;
  register double x3, y3, z3, x4, y4, z4, x5, y5, z5, x6, y6, z6,
    x7, y7, z7, x8, y8, z8, x9, y9, z9, x10, y10, z10,
    x11, y11, z11, x12, y12, z12;
  double r0[2 * MAXL + 1];

  fgamma (r0, t, ltot);
  term = -2.0 * w;
  sterm = 1.0;
  for (m = 0; m <= ltot; ++m)
    {
      r0[m] = sr * sterm * r0[m];
      sterm *= term;
    }
  r[0][0][0] = r0[0];
  if (ltot == 0)
    {
      return;
    }
  x = pq[0];
  y = pq[1];
  z = pq[2];
  r[0][0][1] = z * r0[1];
  r[0][1][0] = y * r0[1];
  r[1][0][0] = x * r0[1];
  if (ltot == 1)
    {
      return;
    }
  x2 = x * x;
  y2 = y * y;
  z2 = z * z;
  r[0][0][2] = z2 * r0[2] + r0[1];
  r[0][1][1] = z * y * r0[2];
  r[0][2][0] = y2 * r0[2] + r0[1];
  r[1][0][1] = z * x * r0[2];
  r[1][1][0] = x * y * r0[2];
  r[2][0][0] = x2 * r0[2] + r0[1];
  if (ltot == 2)
    {
      return;
    }
  x3 = x2 * x;
  y3 = y2 * y;
  z3 = z2 * z;
  r[0][0][3] = z3 * r0[3] + 3.0 * z * r0[2];
  r[0][1][2] = y * z2 * r0[3] + y * r0[2];
  r[1][0][2] = x * z2 * r0[3] + x * r0[2];
  r[0][2][1] = y2 * z * r0[3] + z * r0[2];
  r[1][1][1] = x * y * z * r0[3];
  r[2][0][1] = x2 * z * r0[3] + z * r0[2];
  r[0][3][0] = y3 * r0[3] + 3.0 * y * r0[2];
  r[1][2][0] = x * y2 * r0[3] + x * r0[2];
  r[2][1][0] = x2 * y * r0[3] + y * r0[2];
  r[3][0][0] = x3 * r0[3] + 3.0 * x * r0[2];
  if (ltot == 3)
    {
      return;
    }
  x4 = x3 * x;
  y4 = y3 * y;
  z4 = z3 * z;
  r[0][0][4] = z4 * r0[4] + 6.0 * z2 * r0[3] + 3.0 * r0[2];
  r[0][1][3] = y * z3 * r0[4] + 3.0 * y * z * r0[3];
  r[1][0][3] = x * z3 * r0[4] + 3.0 * x * z * r0[3];
  r[0][2][2] = y2 * z2 * r0[4] + y2 * r0[3] + z2 * r0[3] + r0[2];
  r[1][1][2] = x * y * z2 * r0[4] + x * y * r0[3];
  r[2][0][2] = x2 * z2 * r0[4] + x2 * r0[3] + z2 * r0[3] + r0[2];
  r[0][3][1] = y3 * z * r0[4] + 3.0 * y * z * r0[3];
  r[1][2][1] = x * y2 * z * r0[4] + x * z * r0[3];
  r[2][1][1] = x2 * y * z * r0[4] + y * z * r0[3];
  r[3][0][1] = x3 * z * r0[4] + 3.0 * x * z * r0[3];
  r[0][4][0] = y4 * r0[4] + 6.0 * y2 * r0[3] + 3.0 * r0[2];
  r[1][3][0] = x * y3 * r0[4] + 3.0 * x * y * r0[3];
  r[2][2][0] = x2 * y2 * r0[4] + x2 * r0[3] + y2 * r0[3] + r0[2];
  r[3][1][0] = x3 * y * r0[4] + 3.0 * x * y * r0[3];
  r[4][0][0] = x4 * r0[4] + 6.0 * x2 * r0[3] + 3.0 * r0[2];
  if (ltot == 4)
    {
      return;
    }
  x5 = x4 * x;
  y5 = y4 * y;
  z5 = z4 * z;
  r[0][0][5] = z5 * r0[5] + 10.0 * z3 * r0[4] + 15.0 * z * r0[3];
  r[0][1][4] = y * z4 * r0[5] + 6.0 * y * z2 * r0[4] + 3.0 * y * r0[3];
  r[1][0][4] = x * z4 * r0[5] + 6.0 * x * z2 * r0[4] + 3.0 * x * r0[3];
  r[0][2][3] = y2 * z3 * r0[5] + 3.0 * y2 * z * r0[4] + z3 * r0[4] +
    3.0 * z * r0[3];
  r[1][1][3] = x * y * z3 * r0[5] + 3.0 * x * y * z * r0[4];
  r[2][0][3] = x2 * z3 * r0[5] + 3.0 * x2 * z * r0[4] + z3 * r0[4] +
    3.0 * z * r0[3];
  r[0][3][2] = y3 * z2 * r0[5] + y3 * r0[4] + 3.0 * y * z2 * r0[4] +
    3.0 * y * r0[3];
  r[1][2][2] = x * y2 * z2 * r0[5] + x * y2 * r0[4] + x * z2 * r0[4] +
    x * r0[3];
  r[2][1][2] = x2 * y * z2 * r0[5] + x2 * y * r0[4] + y * z2 * r0[4] +
    y * r0[3];
  r[3][0][2] = x3 * z2 * r0[5] + x3 * r0[4] + 3.0 * x * z2 * r0[4] +
    3.0 * x * r0[3];
  r[0][4][1] = y4 * z * r0[5] + 6.0 * y2 * z * r0[4] + 3.0 * z * r0[3];
  r[1][3][1] = x * y3 * z * r0[5] + 3.0 * x * y * z * r0[4];
  r[2][2][1] = x2 * y2 * z * r0[5] + x2 * z * r0[4] + y2 * z * r0[4] +
    z * r0[3];
  r[3][1][1] = x3 * y * z * r0[5] + 3.0 * x * y * z * r0[4];
  r[4][0][1] = x4 * z * r0[5] + 6.0 * x2 * z * r0[4] + 3.0 * z * r0[3];
  r[0][5][0] = y5 * r0[5] + 10.0 * y3 * r0[4] + 15.0 * y * r0[3];
  r[1][4][0] = x * y4 * r0[5] + 6.0 * x * y2 * r0[4] + 3.0 * x * r0[3];
  r[2][3][0] = x2 * y3 * r0[5] + 3.0 * x2 * y * r0[4] + y3 * r0[4] +
    3.0 * y * r0[3];
  r[3][2][0] = x3 * y2 * r0[5] + x3 * r0[4] + 3.0 * x * y2 * r0[4] +
    3.0 * x * r0[3];
  r[4][1][0] = x4 * y * r0[5] + 6.0 * x2 * y * r0[4] + 3.0 * y * r0[3];
  r[5][0][0] = x5 * r0[5] + 10.0 * x3 * r0[4] + 15.0 * x * r0[3];
  if (ltot == 5)
    {
      return;
    }
  x6 = x5 * x;
  y6 = y5 * y;
  z6 = z5 * z;
  r[0][0][6] =
    z6 * r0[6] + 15.0 * z4 * r0[5] + 45.0 * z2 * r0[4] + 15.0 * r0[3];
  r[0][1][5] = y * z5 * r0[6] + 10.0 * y * z3 * r0[5] + 15.0 * y * z * r0[4];
  r[1][0][5] = x * z5 * r0[6] + 10.0 * x * z3 * r0[5] + 15.0 * x * z * r0[4];
  r[0][2][4] = y2 * z4 * r0[6] + 6.0 * y2 * z2 * r0[5] + 3.0 * y2 * r0[4] +
    z4 * r0[5] + 6.0 * z2 * r0[4] + 3.0 * r0[3];
  r[1][1][4] =
    x * y * z4 * r0[6] + 6.0 * x * y * z2 * r0[5] + 3.0 * x * y * r0[4];
  r[2][0][4] =
    x2 * z4 * r0[6] + 6.0 * x2 * z2 * r0[5] + 3.0 * x2 * r0[4] + z4 * r0[5] +
    6.0 * z2 * r0[4] + 3.0 * r0[3];
  r[0][3][3] =
    y3 * z3 * r0[6] + 3.0 * y3 * z * r0[5] + 3.0 * y * z3 * r0[5] +
    9.0 * y * z * r0[4];
  r[1][2][3] =
    x * y2 * z3 * r0[6] + 3.0 * x * y2 * z * r0[5] + x * z3 * r0[5] +
    3.0 * x * z * r0[4];
  r[2][1][3] =
    x2 * y * z3 * r0[6] + 3.0 * x2 * y * z * r0[5] + y * z3 * r0[5] +
    3.0 * y * z * r0[4];
  r[3][0][3] =
    x3 * z3 * r0[6] + 3.0 * x3 * z * r0[5] + 3.0 * x * z3 * r0[5] +
    9.0 * x * z * r0[4];
  r[0][4][2] =
    y4 * z2 * r0[6] + y4 * r0[5] + 6.0 * y2 * z2 * r0[5] + 6.0 * y2 * r0[4] +
    3.0 * z2 * r0[4] + 3.0 * r0[3];
  r[1][3][2] =
    x * y3 * z2 * r0[6] + x * y3 * r0[5] + 3.0 * x * y * z2 * r0[5] +
    3.0 * x * y * r0[4];
  r[2][2][2] =
    x2 * y2 * z2 * r0[6] + x2 * y2 * r0[5] + x2 * z2 * r0[5] + x2 * r0[4] +
    y2 * z2 * r0[5] + y2 * r0[4] + z2 * r0[4] + r0[3];
  r[3][1][2] =
    x3 * y * z2 * r0[6] + x3 * y * r0[5] + 3.0 * x * y * z2 * r0[5] +
    3.0 * x * y * r0[4];
  r[4][0][2] =
    x4 * z2 * r0[6] + x4 * r0[5] + 6.0 * x2 * z2 * r0[5] + 6.0 * x2 * r0[4] +
    3.0 * z2 * r0[4] + 3.0 * r0[3];
  r[0][5][1] = y5 * z * r0[6] + 10.0 * y3 * z * r0[5] + 15.0 * y * z * r0[4];
  r[1][4][1] =
    x * y4 * z * r0[6] + 6.0 * x * y2 * z * r0[5] + 3.0 * x * z * r0[4];
  r[2][3][1] =
    x2 * y3 * z * r0[6] + 3.0 * x2 * y * z * r0[5] + y3 * z * r0[5] +
    3.0 * y * z * r0[4];
  r[3][2][1] =
    x3 * y2 * z * r0[6] + x3 * z * r0[5] + 3.0 * x * y2 * z * r0[5] +
    3.0 * x * z * r0[4];
  r[4][1][1] =
    x4 * y * z * r0[6] + 6.0 * x2 * y * z * r0[5] + 3.0 * y * z * r0[4];
  r[5][0][1] = x5 * z * r0[6] + 10.0 * x3 * z * r0[5] + 15.0 * x * z * r0[4];
  r[0][6][0] =
    y6 * r0[6] + 15.0 * y4 * r0[5] + 45.0 * y2 * r0[4] + 15.0 * r0[3];
  r[1][5][0] = x * y5 * r0[6] + 10.0 * x * y3 * r0[5] + 15.0 * x * y * r0[4];
  r[2][4][0] = x2 * y4 * r0[6] + 6.0 * x2 * y2 * r0[5] + 3.0 * x2 * r0[4] +
    y4 * r0[5] + 6.0 * y2 * r0[4] + 3.0 * r0[3];
  r[3][3][0] = x3 * y3 * r0[6] + 3.0 * x3 * y * r0[5] + 3.0 * x * y3 * r0[5] +
    9.0 * x * y * r0[4];
  r[4][2][0] = x4 * y2 * r0[6] + x4 * r0[5] + 6.0 * x2 * y2 * r0[5] +
    6.0 * x2 * r0[4] + 3.0 * y2 * r0[4] + 3.0 * r0[3];
  r[5][1][0] = x5 * y * r0[6] + 10.0 * x3 * y * r0[5] + 15.0 * x * y * r0[4];
  r[6][0][0] =
    x6 * r0[6] + 15.0 * x4 * r0[5] + 45.0 * x2 * r0[4] + 15.0 * r0[3];
  if (ltot == 6)
    {
      return;
    }
  x7 = x6 * x;
  y7 = y6 * y;
  z7 = z6 * z;
  r[0][0][7] =
    z7 * r0[7] + 21.0 * z5 * r0[6] + 105.0 * z3 * r0[5] + 105.0 * z * r0[4];
  r[0][1][6] =
    y * z6 * r0[7] + 15.0 * y * z4 * r0[6] + 45.0 * y * z2 * r0[5] +
    15.0 * y * r0[4];
  r[1][0][6] =
    x * z6 * r0[7] + 15.0 * x * z4 * r0[6] + 45.0 * x * z2 * r0[5] +
    15.0 * x * r0[4];
  r[0][2][5] =
    y2 * z5 * r0[7] + 10.0 * y2 * z3 * r0[6] + 15.0 * y2 * z * r0[5] +
    z5 * r0[6] + 10.0 * z3 * r0[5] + 15.0 * z * r0[4];
  r[1][1][5] =
    x * y * z5 * r0[7] + 10.0 * x * y * z3 * r0[6] + 15.0 * x * y * z * r0[5];
  r[2][0][5] =
    x2 * z5 * r0[7] + 10.0 * x2 * z3 * r0[6] + 15.0 * x2 * z * r0[5] +
    z5 * r0[6] + 10.0 * z3 * r0[5] + 15.0 * z * r0[4];
  r[0][3][4] =
    y3 * z4 * r0[7] + 6.0 * y3 * z2 * r0[6] + 3.0 * y3 * r0[5] +
    3.0 * y * z4 * r0[6] + 18.0 * y * z2 * r0[5] + 9.0 * y * r0[4];
  r[1][2][4] =
    x * y2 * z4 * r0[7] + 6.0 * x * y2 * z2 * r0[6] + 3.0 * x * y2 * r0[5] +
    x * z4 * r0[6] + 6.0 * x * z2 * r0[5] + 3.0 * x * r0[4];
  r[2][1][4] =
    x2 * y * z4 * r0[7] + 6.0 * x2 * y * z2 * r0[6] + 3.0 * x2 * y * r0[5] +
    y * z4 * r0[6] + 6.0 * y * z2 * r0[5] + 3.0 * y * r0[4];
  r[3][0][4] =
    x3 * z4 * r0[7] + 6.0 * x3 * z2 * r0[6] + 3.0 * x3 * r0[5] +
    3.0 * x * z4 * r0[6] + 18.0 * x * z2 * r0[5] + 9.0 * x * r0[4];
  r[0][4][3] =
    y4 * z3 * r0[7] + 3.0 * y4 * z * r0[6] + 6.0 * y2 * z3 * r0[6] +
    18.0 * y2 * z * r0[5] + 3.0 * z3 * r0[5] + 9.0 * z * r0[4];
  r[1][3][3] =
    x * y3 * z3 * r0[7] + 3.0 * x * y3 * z * r0[6] +
    3.0 * x * y * z3 * r0[6] + 9.0 * x * y * z * r0[5];
  r[2][2][3] =
    x2 * y2 * z3 * r0[7] + 3.0 * x2 * y2 * z * r0[6] + x2 * z3 * r0[6] +
    3.0 * x2 * z * r0[5] + y2 * z3 * r0[6] + 3.0 * y2 * z * r0[5] +
    z3 * r0[5] + 3.0 * z * r0[4];
  r[3][1][3] =
    x3 * y * z3 * r0[7] + 3.0 * x3 * y * z * r0[6] +
    3.0 * x * y * z3 * r0[6] + 9.0 * x * y * z * r0[5];
  r[4][0][3] =
    x4 * z3 * r0[7] + 3.0 * x4 * z * r0[6] + 6.0 * x2 * z3 * r0[6] +
    18.0 * x2 * z * r0[5] + 3.0 * z3 * r0[5] + 9.0 * z * r0[4];
  r[0][5][2] =
    y5 * z2 * r0[7] + y5 * r0[6] + 10.0 * y3 * z2 * r0[6] +
    10.0 * y3 * r0[5] + 15.0 * y * z2 * r0[5] + 15.0 * y * r0[4];
  r[1][4][2] =
    x * y4 * z2 * r0[7] + x * y4 * r0[6] + 6.0 * x * y2 * z2 * r0[6] +
    6.0 * x * y2 * r0[5] + 3.0 * x * z2 * r0[5] + 3.0 * x * r0[4];
  r[2][3][2] =
    x2 * y3 * z2 * r0[7] + x2 * y3 * r0[6] + 3.0 * x2 * y * z2 * r0[6] +
    3.0 * x2 * y * r0[5] + y3 * z2 * r0[6] + y3 * r0[5] +
    3.0 * y * z2 * r0[5] + 3.0 * y * r0[4];
  r[3][2][2] =
    x3 * y2 * z2 * r0[7] + x3 * y2 * r0[6] + x3 * z2 * r0[6] + x3 * r0[5] +
    3.0 * x * y2 * z2 * r0[6] + 3.0 * x * y2 * r0[5] + 3.0 * x * z2 * r0[5] +
    3.0 * x * r0[4];
  r[4][1][2] =
    x4 * y * z2 * r0[7] + x4 * y * r0[6] + 6.0 * x2 * y * z2 * r0[6] +
    6.0 * x2 * y * r0[5] + 3.0 * y * z2 * r0[5] + 3.0 * y * r0[4];
  r[5][0][2] =
    x5 * z2 * r0[7] + x5 * r0[6] + 10.0 * x3 * z2 * r0[6] +
    10.0 * x3 * r0[5] + 15.0 * x * z2 * r0[5] + 15.0 * x * r0[4];
  r[0][6][1] =
    y6 * z * r0[7] + 15.0 * y4 * z * r0[6] + 45.0 * y2 * z * r0[5] +
    15.0 * z * r0[4];
  r[1][5][1] =
    x * y5 * z * r0[7] + 10.0 * x * y3 * z * r0[6] + 15.0 * x * y * z * r0[5];
  r[2][4][1] =
    x2 * y4 * z * r0[7] + 6.0 * x2 * y2 * z * r0[6] + 3.0 * x2 * z * r0[5] +
    y4 * z * r0[6] + 6.0 * y2 * z * r0[5] + 3.0 * z * r0[4];
  r[3][3][1] =
    x3 * y3 * z * r0[7] + 3.0 * x3 * y * z * r0[6] +
    3.0 * x * y3 * z * r0[6] + 9.0 * x * y * z * r0[5];
  r[4][2][1] =
    x4 * y2 * z * r0[7] + x4 * z * r0[6] + 6.0 * x2 * y2 * z * r0[6] +
    6.0 * x2 * z * r0[5] + 3.0 * y2 * z * r0[5] + 3.0 * z * r0[4];
  r[5][1][1] =
    x5 * y * z * r0[7] + 10.0 * x3 * y * z * r0[6] + 15.0 * x * y * z * r0[5];
  r[6][0][1] =
    x6 * z * r0[7] + 15.0 * x4 * z * r0[6] + 45.0 * x2 * z * r0[5] +
    15.0 * z * r0[4];
  r[0][7][0] =
    y7 * r0[7] + 21.0 * y5 * r0[6] + 105.0 * y3 * r0[5] + 105.0 * y * r0[4];
  r[1][6][0] =
    x * y6 * r0[7] + 15.0 * x * y4 * r0[6] + 45.0 * x * y2 * r0[5] +
    15.0 * x * r0[4];
  r[2][5][0] =
    x2 * y5 * r0[7] + 10.0 * x2 * y3 * r0[6] + 15.0 * x2 * y * r0[5] +
    y5 * r0[6] + 10.0 * y3 * r0[5] + 15.0 * y * r0[4];
  r[3][4][0] =
    x3 * y4 * r0[7] + 6.0 * x3 * y2 * r0[6] + 3.0 * x3 * r0[5] +
    3.0 * x * y4 * r0[6] + 18.0 * x * y2 * r0[5] + 9.0 * x * r0[4];
  r[4][3][0] =
    x4 * y3 * r0[7] + 3.0 * x4 * y * r0[6] + 6.0 * x2 * y3 * r0[6] +
    18.0 * x2 * y * r0[5] + 3.0 * y3 * r0[5] + 9.0 * y * r0[4];
  r[5][2][0] =
    x5 * y2 * r0[7] + x5 * r0[6] + 10.0 * x3 * y2 * r0[6] +
    10.0 * x3 * r0[5] + 15.0 * x * y2 * r0[5] + 15.0 * x * r0[4];
  r[6][1][0] =
    x6 * y * r0[7] + 15.0 * x4 * y * r0[6] + 45.0 * x2 * y * r0[5] +
    15.0 * y * r0[4];
  r[7][0][0] =
    x7 * r0[7] + 21.0 * x5 * r0[6] + 105.0 * x3 * r0[5] + 105.0 * x * r0[4];
  if (ltot == 7)
    {
      return;
    }
  x8 = x7 * x;
  y8 = y7 * y;
  z8 = z7 * z;
  r[0][0][8] =
    z8 * r0[8] + 28.0 * z6 * r0[7] + 210.0 * z4 * r0[6] + 420.0 * z2 * r0[5] +
    105.0 * r0[4];
  r[0][1][7] =
    y * z7 * r0[8] + 21.0 * y * z5 * r0[7] + 105.0 * y * z3 * r0[6] +
    105.0 * y * z * r0[5];
  r[1][0][7] =
    x * z7 * r0[8] + 21.0 * x * z5 * r0[7] + 105.0 * x * z3 * r0[6] +
    105.0 * x * z * r0[5];
  r[0][2][6] =
    y2 * z6 * r0[8] + 15.0 * y2 * z4 * r0[7] + 45.0 * y2 * z2 * r0[6] +
    15.0 * y2 * r0[5] + z6 * r0[7] + 15.0 * z4 * r0[6] + 45.0 * z2 * r0[5] +
    15.0 * r0[4];
  r[1][1][6] =
    x * y * z6 * r0[8] + 15.0 * x * y * z4 * r0[7] +
    45.0 * x * y * z2 * r0[6] + 15.0 * x * y * r0[5];
  r[2][0][6] =
    x2 * z6 * r0[8] + 15.0 * x2 * z4 * r0[7] + 45.0 * x2 * z2 * r0[6] +
    15.0 * x2 * r0[5] + z6 * r0[7] + 15.0 * z4 * r0[6] + 45.0 * z2 * r0[5] +
    15.0 * r0[4];
  r[0][3][5] =
    y3 * z5 * r0[8] + 10.0 * y3 * z3 * r0[7] + 15.0 * y3 * z * r0[6] +
    3.0 * y * z5 * r0[7] + 30.0 * y * z3 * r0[6] + 45.0 * y * z * r0[5];
  r[1][2][5] =
    x * y2 * z5 * r0[8] + 10.0 * x * y2 * z3 * r0[7] +
    15.0 * x * y2 * z * r0[6] + x * z5 * r0[7] + 10.0 * x * z3 * r0[6] +
    15.0 * x * z * r0[5];
  r[2][1][5] =
    x2 * y * z5 * r0[8] + 10.0 * x2 * y * z3 * r0[7] +
    15.0 * x2 * y * z * r0[6] + y * z5 * r0[7] + 10.0 * y * z3 * r0[6] +
    15.0 * y * z * r0[5];
  r[3][0][5] =
    x3 * z5 * r0[8] + 10.0 * x3 * z3 * r0[7] + 15.0 * x3 * z * r0[6] +
    3.0 * x * z5 * r0[7] + 30.0 * x * z3 * r0[6] + 45.0 * x * z * r0[5];
  r[0][4][4] =
    y4 * z4 * r0[8] + 6.0 * y4 * z2 * r0[7] + 3.0 * y4 * r0[6] +
    6.0 * y2 * z4 * r0[7] + 36.0 * y2 * z2 * r0[6] + 18.0 * y2 * r0[5] +
    3.0 * z4 * r0[6] + 18.0 * z2 * r0[5] + 9.0 * r0[4];
  r[1][3][4] =
    x * y3 * z4 * r0[8] + 6.0 * x * y3 * z2 * r0[7] + 3.0 * x * y3 * r0[6] +
    3.0 * x * y * z4 * r0[7] + 18.0 * x * y * z2 * r0[6] +
    9.0 * x * y * r0[5];
  r[2][2][4] =
    x2 * y2 * z4 * r0[8] + 6.0 * x2 * y2 * z2 * r0[7] +
    3.0 * x2 * y2 * r0[6] + x2 * z4 * r0[7] + 6.0 * x2 * z2 * r0[6] +
    3.0 * x2 * r0[5] + y2 * z4 * r0[7] + 6.0 * y2 * z2 * r0[6] +
    3.0 * y2 * r0[5] + z4 * r0[6] + 6.0 * z2 * r0[5] + 3.0 * r0[4];
  r[3][1][4] =
    x3 * y * z4 * r0[8] + 6.0 * x3 * y * z2 * r0[7] + 3.0 * x3 * y * r0[6] +
    3.0 * x * y * z4 * r0[7] + 18.0 * x * y * z2 * r0[6] +
    9.0 * x * y * r0[5];
  r[4][0][4] =
    x4 * z4 * r0[8] + 6.0 * x4 * z2 * r0[7] + 3.0 * x4 * r0[6] +
    6.0 * x2 * z4 * r0[7] + 36.0 * x2 * z2 * r0[6] + 18.0 * x2 * r0[5] +
    3.0 * z4 * r0[6] + 18.0 * z2 * r0[5] + 9.0 * r0[4];
  r[0][5][3] =
    y5 * z3 * r0[8] + 3.0 * y5 * z * r0[7] + 10.0 * y3 * z3 * r0[7] +
    30.0 * y3 * z * r0[6] + 15.0 * y * z3 * r0[6] + 45.0 * y * z * r0[5];
  r[1][4][3] =
    x * y4 * z3 * r0[8] + 3.0 * x * y4 * z * r0[7] +
    6.0 * x * y2 * z3 * r0[7] + 18.0 * x * y2 * z * r0[6] +
    3.0 * x * z3 * r0[6] + 9.0 * x * z * r0[5];
  r[2][3][3] =
    x2 * y3 * z3 * r0[8] + 3.0 * x2 * y3 * z * r0[7] +
    3.0 * x2 * y * z3 * r0[7] + 9.0 * x2 * y * z * r0[6] + y3 * z3 * r0[7] +
    3.0 * y3 * z * r0[6] + 3.0 * y * z3 * r0[6] + 9.0 * y * z * r0[5];
  r[3][2][3] =
    x3 * y2 * z3 * r0[8] + 3.0 * x3 * y2 * z * r0[7] + x3 * z3 * r0[7] +
    3.0 * x3 * z * r0[6] + 3.0 * x * y2 * z3 * r0[7] +
    9.0 * x * y2 * z * r0[6] + 3.0 * x * z3 * r0[6] + 9.0 * x * z * r0[5];
  r[4][1][3] =
    x4 * y * z3 * r0[8] + 3.0 * x4 * y * z * r0[7] +
    6.0 * x2 * y * z3 * r0[7] + 18.0 * x2 * y * z * r0[6] +
    3.0 * y * z3 * r0[6] + 9.0 * y * z * r0[5];
  r[5][0][3] =
    x5 * z3 * r0[8] + 3.0 * x5 * z * r0[7] + 10.0 * x3 * z3 * r0[7] +
    30.0 * x3 * z * r0[6] + 15.0 * x * z3 * r0[6] + 45.0 * x * z * r0[5];
  r[0][6][2] =
    y6 * z2 * r0[8] + y6 * r0[7] + 15.0 * y4 * z2 * r0[7] +
    15.0 * y4 * r0[6] + 45.0 * y2 * z2 * r0[6] + 45.0 * y2 * r0[5] +
    15.0 * z2 * r0[5] + 15.0 * r0[4];
  r[1][5][2] =
    x * y5 * z2 * r0[8] + x * y5 * r0[7] + 10.0 * x * y3 * z2 * r0[7] +
    10.0 * x * y3 * r0[6] + 15.0 * x * y * z2 * r0[6] + 15.0 * x * y * r0[5];
  r[2][4][2] =
    x2 * y4 * z2 * r0[8] + x2 * y4 * r0[7] + 6.0 * x2 * y2 * z2 * r0[7] +
    6.0 * x2 * y2 * r0[6] + 3.0 * x2 * z2 * r0[6] + 3.0 * x2 * r0[5] +
    y4 * z2 * r0[7] + y4 * r0[6] + 6.0 * y2 * z2 * r0[6] + 6.0 * y2 * r0[5] +
    3.0 * z2 * r0[5] + 3.0 * r0[4];
  r[3][3][2] =
    x3 * y3 * z2 * r0[8] + x3 * y3 * r0[7] + 3.0 * x3 * y * z2 * r0[7] +
    3.0 * x3 * y * r0[6] + 3.0 * x * y3 * z2 * r0[7] + 3.0 * x * y3 * r0[6] +
    9.0 * x * y * z2 * r0[6] + 9.0 * x * y * r0[5];
  r[4][2][2] =
    x4 * y2 * z2 * r0[8] + x4 * y2 * r0[7] + x4 * z2 * r0[7] + x4 * r0[6] +
    6.0 * x2 * y2 * z2 * r0[7] + 6.0 * x2 * y2 * r0[6] +
    6.0 * x2 * z2 * r0[6] + 6.0 * x2 * r0[5] + 3.0 * y2 * z2 * r0[6] +
    3.0 * y2 * r0[5] + 3.0 * z2 * r0[5] + 3.0 * r0[4];
  r[5][1][2] =
    x5 * y * z2 * r0[8] + x5 * y * r0[7] + 10.0 * x3 * y * z2 * r0[7] +
    10.0 * x3 * y * r0[6] + 15.0 * x * y * z2 * r0[6] + 15.0 * x * y * r0[5];
  r[6][0][2] =
    x6 * z2 * r0[8] + x6 * r0[7] + 15.0 * x4 * z2 * r0[7] +
    15.0 * x4 * r0[6] + 45.0 * x2 * z2 * r0[6] + 45.0 * x2 * r0[5] +
    15.0 * z2 * r0[5] + 15.0 * r0[4];
  r[0][7][1] =
    y7 * z * r0[8] + 21.0 * y5 * z * r0[7] + 105.0 * y3 * z * r0[6] +
    105.0 * y * z * r0[5];
  r[1][6][1] =
    x * y6 * z * r0[8] + 15.0 * x * y4 * z * r0[7] +
    45.0 * x * y2 * z * r0[6] + 15.0 * x * z * r0[5];
  r[2][5][1] =
    x2 * y5 * z * r0[8] + 10.0 * x2 * y3 * z * r0[7] +
    15.0 * x2 * y * z * r0[6] + y5 * z * r0[7] + 10.0 * y3 * z * r0[6] +
    15.0 * y * z * r0[5];
  r[3][4][1] =
    x3 * y4 * z * r0[8] + 6.0 * x3 * y2 * z * r0[7] + 3.0 * x3 * z * r0[6] +
    3.0 * x * y4 * z * r0[7] + 18.0 * x * y2 * z * r0[6] +
    9.0 * x * z * r0[5];
  r[4][3][1] =
    x4 * y3 * z * r0[8] + 3.0 * x4 * y * z * r0[7] +
    6.0 * x2 * y3 * z * r0[7] + 18.0 * x2 * y * z * r0[6] +
    3.0 * y3 * z * r0[6] + 9.0 * y * z * r0[5];
  r[5][2][1] =
    x5 * y2 * z * r0[8] + x5 * z * r0[7] + 10.0 * x3 * y2 * z * r0[7] +
    10.0 * x3 * z * r0[6] + 15.0 * x * y2 * z * r0[6] + 15.0 * x * z * r0[5];
  r[6][1][1] =
    x6 * y * z * r0[8] + 15.0 * x4 * y * z * r0[7] +
    45.0 * x2 * y * z * r0[6] + 15.0 * y * z * r0[5];
  r[7][0][1] =
    x7 * z * r0[8] + 21.0 * x5 * z * r0[7] + 105.0 * x3 * z * r0[6] +
    105.0 * x * z * r0[5];
  r[0][8][0] =
    y8 * r0[8] + 28.0 * y6 * r0[7] + 210.0 * y4 * r0[6] + 420.0 * y2 * r0[5] +
    105.0 * r0[4];
  r[1][7][0] =
    x * y7 * r0[8] + 21.0 * x * y5 * r0[7] + 105.0 * x * y3 * r0[6] +
    105.0 * x * y * r0[5];
  r[2][6][0] =
    x2 * y6 * r0[8] + 15.0 * x2 * y4 * r0[7] + 45.0 * x2 * y2 * r0[6] +
    15.0 * x2 * r0[5] + y6 * r0[7] + 15.0 * y4 * r0[6] + 45.0 * y2 * r0[5] +
    15.0 * r0[4];
  r[3][5][0] =
    x3 * y5 * r0[8] + 10.0 * x3 * y3 * r0[7] + 15.0 * x3 * y * r0[6] +
    3.0 * x * y5 * r0[7] + 30.0 * x * y3 * r0[6] + 45.0 * x * y * r0[5];
  r[4][4][0] =
    x4 * y4 * r0[8] + 6.0 * x4 * y2 * r0[7] + 3.0 * x4 * r0[6] +
    6.0 * x2 * y4 * r0[7] + 36.0 * x2 * y2 * r0[6] + 18.0 * x2 * r0[5] +
    3.0 * y4 * r0[6] + 18.0 * y2 * r0[5] + 9.0 * r0[4];
  r[5][3][0] =
    x5 * y3 * r0[8] + 3.0 * x5 * y * r0[7] + 10.0 * x3 * y3 * r0[7] +
    30.0 * x3 * y * r0[6] + 15.0 * x * y3 * r0[6] + 45.0 * x * y * r0[5];
  r[6][2][0] =
    x6 * y2 * r0[8] + x6 * r0[7] + 15.0 * x4 * y2 * r0[7] +
    15.0 * x4 * r0[6] + 45.0 * x2 * y2 * r0[6] + 45.0 * x2 * r0[5] +
    15.0 * y2 * r0[5] + 15.0 * r0[4];
  r[7][1][0] =
    x7 * y * r0[8] + 21.0 * x5 * y * r0[7] + 105.0 * x3 * y * r0[6] +
    105.0 * x * y * r0[5];
  r[8][0][0] =
    x8 * r0[8] + 28.0 * x6 * r0[7] + 210.0 * x4 * r0[6] + 420.0 * x2 * r0[5] +
    105.0 * r0[4];
  if (ltot == 8)
    {
      return;
    }
  x9 = x8 * x;
  y9 = y8 * y;
  z9 = z8 * z;
  r[0][0][9] =
    z9 * r0[9] + 36.0 * z7 * r0[8] + 378.0 * z5 * r0[7] +
    1260.0 * z3 * r0[6] + 945.0 * z * r0[5];
  r[0][1][8] =
    y * z8 * r0[9] + 28.0 * y * z6 * r0[8] + 210.0 * y * z4 * r0[7] +
    420.0 * y * z2 * r0[6] + 105.0 * y * r0[5];
  r[1][0][8] =
    x * z8 * r0[9] + 28.0 * x * z6 * r0[8] + 210.0 * x * z4 * r0[7] +
    420.0 * x * z2 * r0[6] + 105.0 * x * r0[5];
  r[0][2][7] =
    y2 * z7 * r0[9] + 21.0 * y2 * z5 * r0[8] + 105.0 * y2 * z3 * r0[7] +
    105.0 * y2 * z * r0[6] + z7 * r0[8] + 21.0 * z5 * r0[7] +
    105.0 * z3 * r0[6] + 105.0 * z * r0[5];
  r[1][1][7] =
    x * y * z7 * r0[9] + 21.0 * x * y * z5 * r0[8] +
    105.0 * x * y * z3 * r0[7] + 105.0 * x * y * z * r0[6];
  r[2][0][7] =
    x2 * z7 * r0[9] + 21.0 * x2 * z5 * r0[8] + 105.0 * x2 * z3 * r0[7] +
    105.0 * x2 * z * r0[6] + z7 * r0[8] + 21.0 * z5 * r0[7] +
    105.0 * z3 * r0[6] + 105.0 * z * r0[5];
  r[0][3][6] =
    y3 * z6 * r0[9] + 15.0 * y3 * z4 * r0[8] + 45.0 * y3 * z2 * r0[7] +
    15.0 * y3 * r0[6] + 3.0 * y * z6 * r0[8] + 45.0 * y * z4 * r0[7] +
    135.0 * y * z2 * r0[6] + 45.0 * y * r0[5];
  r[1][2][6] =
    x * y2 * z6 * r0[9] + 15.0 * x * y2 * z4 * r0[8] +
    45.0 * x * y2 * z2 * r0[7] + 15.0 * x * y2 * r0[6] + x * z6 * r0[8] +
    15.0 * x * z4 * r0[7] + 45.0 * x * z2 * r0[6] + 15.0 * x * r0[5];
  r[2][1][6] =
    x2 * y * z6 * r0[9] + 15.0 * x2 * y * z4 * r0[8] +
    45.0 * x2 * y * z2 * r0[7] + 15.0 * x2 * y * r0[6] + y * z6 * r0[8] +
    15.0 * y * z4 * r0[7] + 45.0 * y * z2 * r0[6] + 15.0 * y * r0[5];
  r[3][0][6] =
    x3 * z6 * r0[9] + 15.0 * x3 * z4 * r0[8] + 45.0 * x3 * z2 * r0[7] +
    15.0 * x3 * r0[6] + 3.0 * x * z6 * r0[8] + 45.0 * x * z4 * r0[7] +
    135.0 * x * z2 * r0[6] + 45.0 * x * r0[5];
  r[0][4][5] =
    y4 * z5 * r0[9] + 10.0 * y4 * z3 * r0[8] + 15.0 * y4 * z * r0[7] +
    6.0 * y2 * z5 * r0[8] + 60.0 * y2 * z3 * r0[7] + 90.0 * y2 * z * r0[6] +
    3.0 * z5 * r0[7] + 30.0 * z3 * r0[6] + 45.0 * z * r0[5];
  r[1][3][5] =
    x * y3 * z5 * r0[9] + 10.0 * x * y3 * z3 * r0[8] +
    15.0 * x * y3 * z * r0[7] + 3.0 * x * y * z5 * r0[8] +
    30.0 * x * y * z3 * r0[7] + 45.0 * x * y * z * r0[6];
  r[2][2][5] =
    x2 * y2 * z5 * r0[9] + 10.0 * x2 * y2 * z3 * r0[8] +
    15.0 * x2 * y2 * z * r0[7] + x2 * z5 * r0[8] + 10.0 * x2 * z3 * r0[7] +
    15.0 * x2 * z * r0[6] + y2 * z5 * r0[8] + 10.0 * y2 * z3 * r0[7] +
    15.0 * y2 * z * r0[6] + z5 * r0[7] + 10.0 * z3 * r0[6] + 15.0 * z * r0[5];
  r[3][1][5] =
    x3 * y * z5 * r0[9] + 10.0 * x3 * y * z3 * r0[8] +
    15.0 * x3 * y * z * r0[7] + 3.0 * x * y * z5 * r0[8] +
    30.0 * x * y * z3 * r0[7] + 45.0 * x * y * z * r0[6];
  r[4][0][5] =
    x4 * z5 * r0[9] + 10.0 * x4 * z3 * r0[8] + 15.0 * x4 * z * r0[7] +
    6.0 * x2 * z5 * r0[8] + 60.0 * x2 * z3 * r0[7] + 90.0 * x2 * z * r0[6] +
    3.0 * z5 * r0[7] + 30.0 * z3 * r0[6] + 45.0 * z * r0[5];
  r[0][5][4] =
    y5 * z4 * r0[9] + 6.0 * y5 * z2 * r0[8] + 3.0 * y5 * r0[7] +
    10.0 * y3 * z4 * r0[8] + 60.0 * y3 * z2 * r0[7] + 30.0 * y3 * r0[6] +
    15.0 * y * z4 * r0[7] + 90.0 * y * z2 * r0[6] + 45.0 * y * r0[5];
  r[1][4][4] =
    x * y4 * z4 * r0[9] + 6.0 * x * y4 * z2 * r0[8] + 3.0 * x * y4 * r0[7] +
    6.0 * x * y2 * z4 * r0[8] + 36.0 * x * y2 * z2 * r0[7] +
    18.0 * x * y2 * r0[6] + 3.0 * x * z4 * r0[7] + 18.0 * x * z2 * r0[6] +
    9.0 * x * r0[5];
  r[2][3][4] =
    x2 * y3 * z4 * r0[9] + 6.0 * x2 * y3 * z2 * r0[8] +
    3.0 * x2 * y3 * r0[7] + 3.0 * x2 * y * z4 * r0[8] +
    18.0 * x2 * y * z2 * r0[7] + 9.0 * x2 * y * r0[6] + y3 * z4 * r0[8] +
    6.0 * y3 * z2 * r0[7] + 3.0 * y3 * r0[6] + 3.0 * y * z4 * r0[7] +
    18.0 * y * z2 * r0[6] + 9.0 * y * r0[5];
  r[3][2][4] =
    x3 * y2 * z4 * r0[9] + 6.0 * x3 * y2 * z2 * r0[8] +
    3.0 * x3 * y2 * r0[7] + x3 * z4 * r0[8] + 6.0 * x3 * z2 * r0[7] +
    3.0 * x3 * r0[6] + 3.0 * x * y2 * z4 * r0[8] +
    18.0 * x * y2 * z2 * r0[7] + 9.0 * x * y2 * r0[6] + 3.0 * x * z4 * r0[7] +
    18.0 * x * z2 * r0[6] + 9.0 * x * r0[5];
  r[4][1][4] =
    x4 * y * z4 * r0[9] + 6.0 * x4 * y * z2 * r0[8] + 3.0 * x4 * y * r0[7] +
    6.0 * x2 * y * z4 * r0[8] + 36.0 * x2 * y * z2 * r0[7] +
    18.0 * x2 * y * r0[6] + 3.0 * y * z4 * r0[7] + 18.0 * y * z2 * r0[6] +
    9.0 * y * r0[5];
  r[5][0][4] =
    x5 * z4 * r0[9] + 6.0 * x5 * z2 * r0[8] + 3.0 * x5 * r0[7] +
    10.0 * x3 * z4 * r0[8] + 60.0 * x3 * z2 * r0[7] + 30.0 * x3 * r0[6] +
    15.0 * x * z4 * r0[7] + 90.0 * x * z2 * r0[6] + 45.0 * x * r0[5];
  r[0][6][3] =
    y6 * z3 * r0[9] + 3.0 * y6 * z * r0[8] + 15.0 * y4 * z3 * r0[8] +
    45.0 * y4 * z * r0[7] + 45.0 * y2 * z3 * r0[7] + 135.0 * y2 * z * r0[6] +
    15.0 * z3 * r0[6] + 45.0 * z * r0[5];
  r[1][5][3] =
    x * y5 * z3 * r0[9] + 3.0 * x * y5 * z * r0[8] +
    10.0 * x * y3 * z3 * r0[8] + 30.0 * x * y3 * z * r0[7] +
    15.0 * x * y * z3 * r0[7] + 45.0 * x * y * z * r0[6];
  r[2][4][3] =
    x2 * y4 * z3 * r0[9] + 3.0 * x2 * y4 * z * r0[8] +
    6.0 * x2 * y2 * z3 * r0[8] + 18.0 * x2 * y2 * z * r0[7] +
    3.0 * x2 * z3 * r0[7] + 9.0 * x2 * z * r0[6] + y4 * z3 * r0[8] +
    3.0 * y4 * z * r0[7] + 6.0 * y2 * z3 * r0[7] + 18.0 * y2 * z * r0[6] +
    3.0 * z3 * r0[6] + 9.0 * z * r0[5];
  r[3][3][3] =
    x3 * y3 * z3 * r0[9] + 3.0 * x3 * y3 * z * r0[8] +
    3.0 * x3 * y * z3 * r0[8] + 9.0 * x3 * y * z * r0[7] +
    3.0 * x * y3 * z3 * r0[8] + 9.0 * x * y3 * z * r0[7] +
    9.0 * x * y * z3 * r0[7] + 27.0 * x * y * z * r0[6];
  r[4][2][3] =
    x4 * y2 * z3 * r0[9] + 3.0 * x4 * y2 * z * r0[8] + x4 * z3 * r0[8] +
    3.0 * x4 * z * r0[7] + 6.0 * x2 * y2 * z3 * r0[8] +
    18.0 * x2 * y2 * z * r0[7] + 6.0 * x2 * z3 * r0[7] +
    18.0 * x2 * z * r0[6] + 3.0 * y2 * z3 * r0[7] + 9.0 * y2 * z * r0[6] +
    3.0 * z3 * r0[6] + 9.0 * z * r0[5];
  r[5][1][3] =
    x5 * y * z3 * r0[9] + 3.0 * x5 * y * z * r0[8] +
    10.0 * x3 * y * z3 * r0[8] + 30.0 * x3 * y * z * r0[7] +
    15.0 * x * y * z3 * r0[7] + 45.0 * x * y * z * r0[6];
  r[6][0][3] =
    x6 * z3 * r0[9] + 3.0 * x6 * z * r0[8] + 15.0 * x4 * z3 * r0[8] +
    45.0 * x4 * z * r0[7] + 45.0 * x2 * z3 * r0[7] + 135.0 * x2 * z * r0[6] +
    15.0 * z3 * r0[6] + 45.0 * z * r0[5];
  r[0][7][2] =
    y7 * z2 * r0[9] + y7 * r0[8] + 21.0 * y5 * z2 * r0[8] +
    21.0 * y5 * r0[7] + 105.0 * y3 * z2 * r0[7] + 105.0 * y3 * r0[6] +
    105.0 * y * z2 * r0[6] + 105.0 * y * r0[5];
  r[1][6][2] =
    x * y6 * z2 * r0[9] + x * y6 * r0[8] + 15.0 * x * y4 * z2 * r0[8] +
    15.0 * x * y4 * r0[7] + 45.0 * x * y2 * z2 * r0[7] +
    45.0 * x * y2 * r0[6] + 15.0 * x * z2 * r0[6] + 15.0 * x * r0[5];
  r[2][5][2] =
    x2 * y5 * z2 * r0[9] + x2 * y5 * r0[8] + 10.0 * x2 * y3 * z2 * r0[8] +
    10.0 * x2 * y3 * r0[7] + 15.0 * x2 * y * z2 * r0[7] +
    15.0 * x2 * y * r0[6] + y5 * z2 * r0[8] + y5 * r0[7] +
    10.0 * y3 * z2 * r0[7] + 10.0 * y3 * r0[6] + 15.0 * y * z2 * r0[6] +
    15.0 * y * r0[5];
  r[3][4][2] =
    x3 * y4 * z2 * r0[9] + x3 * y4 * r0[8] + 6.0 * x3 * y2 * z2 * r0[8] +
    6.0 * x3 * y2 * r0[7] + 3.0 * x3 * z2 * r0[7] + 3.0 * x3 * r0[6] +
    3.0 * x * y4 * z2 * r0[8] + 3.0 * x * y4 * r0[7] +
    18.0 * x * y2 * z2 * r0[7] + 18.0 * x * y2 * r0[6] +
    9.0 * x * z2 * r0[6] + 9.0 * x * r0[5];
  r[4][3][2] =
    x4 * y3 * z2 * r0[9] + x4 * y3 * r0[8] + 3.0 * x4 * y * z2 * r0[8] +
    3.0 * x4 * y * r0[7] + 6.0 * x2 * y3 * z2 * r0[8] +
    6.0 * x2 * y3 * r0[7] + 18.0 * x2 * y * z2 * r0[7] +
    18.0 * x2 * y * r0[6] + 3.0 * y3 * z2 * r0[7] + 3.0 * y3 * r0[6] +
    9.0 * y * z2 * r0[6] + 9.0 * y * r0[5];
  r[5][2][2] =
    x5 * y2 * z2 * r0[9] + x5 * y2 * r0[8] + x5 * z2 * r0[8] + x5 * r0[7] +
    10.0 * x3 * y2 * z2 * r0[8] + 10.0 * x3 * y2 * r0[7] +
    10.0 * x3 * z2 * r0[7] + 10.0 * x3 * r0[6] + 15.0 * x * y2 * z2 * r0[7] +
    15.0 * x * y2 * r0[6] + 15.0 * x * z2 * r0[6] + 15.0 * x * r0[5];
  r[6][1][2] =
    x6 * y * z2 * r0[9] + x6 * y * r0[8] + 15.0 * x4 * y * z2 * r0[8] +
    15.0 * x4 * y * r0[7] + 45.0 * x2 * y * z2 * r0[7] +
    45.0 * x2 * y * r0[6] + 15.0 * y * z2 * r0[6] + 15.0 * y * r0[5];
  r[7][0][2] =
    x7 * z2 * r0[9] + x7 * r0[8] + 21.0 * x5 * z2 * r0[8] +
    21.0 * x5 * r0[7] + 105.0 * x3 * z2 * r0[7] + 105.0 * x3 * r0[6] +
    105.0 * x * z2 * r0[6] + 105.0 * x * r0[5];
  r[0][8][1] =
    y8 * z * r0[9] + 28.0 * y6 * z * r0[8] + 210.0 * y4 * z * r0[7] +
    420.0 * y2 * z * r0[6] + 105.0 * z * r0[5];
  r[1][7][1] =
    x * y7 * z * r0[9] + 21.0 * x * y5 * z * r0[8] +
    105.0 * x * y3 * z * r0[7] + 105.0 * x * y * z * r0[6];
  r[2][6][1] =
    x2 * y6 * z * r0[9] + 15.0 * x2 * y4 * z * r0[8] +
    45.0 * x2 * y2 * z * r0[7] + 15.0 * x2 * z * r0[6] + y6 * z * r0[8] +
    15.0 * y4 * z * r0[7] + 45.0 * y2 * z * r0[6] + 15.0 * z * r0[5];
  r[3][5][1] =
    x3 * y5 * z * r0[9] + 10.0 * x3 * y3 * z * r0[8] +
    15.0 * x3 * y * z * r0[7] + 3.0 * x * y5 * z * r0[8] +
    30.0 * x * y3 * z * r0[7] + 45.0 * x * y * z * r0[6];
  r[4][4][1] =
    x4 * y4 * z * r0[9] + 6.0 * x4 * y2 * z * r0[8] + 3.0 * x4 * z * r0[7] +
    6.0 * x2 * y4 * z * r0[8] + 36.0 * x2 * y2 * z * r0[7] +
    18.0 * x2 * z * r0[6] + 3.0 * y4 * z * r0[7] + 18.0 * y2 * z * r0[6] +
    9.0 * z * r0[5];
  r[5][3][1] =
    x5 * y3 * z * r0[9] + 3.0 * x5 * y * z * r0[8] +
    10.0 * x3 * y3 * z * r0[8] + 30.0 * x3 * y * z * r0[7] +
    15.0 * x * y3 * z * r0[7] + 45.0 * x * y * z * r0[6];
  r[6][2][1] =
    x6 * y2 * z * r0[9] + x6 * z * r0[8] + 15.0 * x4 * y2 * z * r0[8] +
    15.0 * x4 * z * r0[7] + 45.0 * x2 * y2 * z * r0[7] +
    45.0 * x2 * z * r0[6] + 15.0 * y2 * z * r0[6] + 15.0 * z * r0[5];
  r[7][1][1] =
    x7 * y * z * r0[9] + 21.0 * x5 * y * z * r0[8] +
    105.0 * x3 * y * z * r0[7] + 105.0 * x * y * z * r0[6];
  r[8][0][1] =
    x8 * z * r0[9] + 28.0 * x6 * z * r0[8] + 210.0 * x4 * z * r0[7] +
    420.0 * x2 * z * r0[6] + 105.0 * z * r0[5];
  r[0][9][0] =
    y9 * r0[9] + 36.0 * y7 * r0[8] + 378.0 * y5 * r0[7] +
    1260.0 * y3 * r0[6] + 945.0 * y * r0[5];
  r[1][8][0] =
    x * y8 * r0[9] + 28.0 * x * y6 * r0[8] + 210.0 * x * y4 * r0[7] +
    420.0 * x * y2 * r0[6] + 105.0 * x * r0[5];
  r[2][7][0] =
    x2 * y7 * r0[9] + 21.0 * x2 * y5 * r0[8] + 105.0 * x2 * y3 * r0[7] +
    105.0 * x2 * y * r0[6] + y7 * r0[8] + 21.0 * y5 * r0[7] +
    105.0 * y3 * r0[6] + 105.0 * y * r0[5];
  r[3][6][0] =
    x3 * y6 * r0[9] + 15.0 * x3 * y4 * r0[8] + 45.0 * x3 * y2 * r0[7] +
    15.0 * x3 * r0[6] + 3.0 * x * y6 * r0[8] + 45.0 * x * y4 * r0[7] +
    135.0 * x * y2 * r0[6] + 45.0 * x * r0[5];
  r[4][5][0] =
    x4 * y5 * r0[9] + 10.0 * x4 * y3 * r0[8] + 15.0 * x4 * y * r0[7] +
    6.0 * x2 * y5 * r0[8] + 60.0 * x2 * y3 * r0[7] + 90.0 * x2 * y * r0[6] +
    3.0 * y5 * r0[7] + 30.0 * y3 * r0[6] + 45.0 * y * r0[5];
  r[5][4][0] =
    x5 * y4 * r0[9] + 6.0 * x5 * y2 * r0[8] + 3.0 * x5 * r0[7] +
    10.0 * x3 * y4 * r0[8] + 60.0 * x3 * y2 * r0[7] + 30.0 * x3 * r0[6] +
    15.0 * x * y4 * r0[7] + 90.0 * x * y2 * r0[6] + 45.0 * x * r0[5];
  r[6][3][0] =
    x6 * y3 * r0[9] + 3.0 * x6 * y * r0[8] + 15.0 * x4 * y3 * r0[8] +
    45.0 * x4 * y * r0[7] + 45.0 * x2 * y3 * r0[7] + 135.0 * x2 * y * r0[6] +
    15.0 * y3 * r0[6] + 45.0 * y * r0[5];
  r[7][2][0] =
    x7 * y2 * r0[9] + x7 * r0[8] + 21.0 * x5 * y2 * r0[8] +
    21.0 * x5 * r0[7] + 105.0 * x3 * y2 * r0[7] + 105.0 * x3 * r0[6] +
    105.0 * x * y2 * r0[6] + 105.0 * x * r0[5];
  r[8][1][0] =
    x8 * y * r0[9] + 28.0 * x6 * y * r0[8] + 210.0 * x4 * y * r0[7] +
    420.0 * x2 * y * r0[6] + 105.0 * y * r0[5];
  r[9][0][0] =
    x9 * r0[9] + 36.0 * x7 * r0[8] + 378.0 * x5 * r0[7] +
    1260.0 * x3 * r0[6] + 945.0 * x * r0[5];
  if (ltot == 9)
    {
      return;
    }
  x10 = x9 * x;
  y10 = y9 * y;
  z10 = z9 * z;
  r[0][0][10] =
    z10 * r0[10] + 45.0 * z8 * r0[9] + 630.0 * z6 * r0[8] +
    3150.0 * z4 * r0[7] + 4725.0 * z2 * r0[6] + 945.0 * r0[5];
  r[0][1][9] =
    y * z9 * r0[10] + 36.0 * y * z7 * r0[9] + 378.0 * y * z5 * r0[8] +
    1260.0 * y * z3 * r0[7] + 945.0 * y * z * r0[6];
  r[1][0][9] =
    x * z9 * r0[10] + 36.0 * x * z7 * r0[9] + 378.0 * x * z5 * r0[8] +
    1260.0 * x * z3 * r0[7] + 945.0 * x * z * r0[6];
  r[0][2][8] =
    y2 * z8 * r0[10] + 28.0 * y2 * z6 * r0[9] + 210.0 * y2 * z4 * r0[8] +
    420.0 * y2 * z2 * r0[7] + 105.0 * y2 * r0[6] + z8 * r0[9] +
    28.0 * z6 * r0[8] + 210.0 * z4 * r0[7] + 420.0 * z2 * r0[6] +
    105.0 * r0[5];
  r[1][1][8] =
    x * y * z8 * r0[10] + 28.0 * x * y * z6 * r0[9] +
    210.0 * x * y * z4 * r0[8] + 420.0 * x * y * z2 * r0[7] +
    105.0 * x * y * r0[6];
  r[2][0][8] =
    x2 * z8 * r0[10] + 28.0 * x2 * z6 * r0[9] + 210.0 * x2 * z4 * r0[8] +
    420.0 * x2 * z2 * r0[7] + 105.0 * x2 * r0[6] + z8 * r0[9] +
    28.0 * z6 * r0[8] + 210.0 * z4 * r0[7] + 420.0 * z2 * r0[6] +
    105.0 * r0[5];
  r[0][3][7] =
    y3 * z7 * r0[10] + 21.0 * y3 * z5 * r0[9] + 105.0 * y3 * z3 * r0[8] +
    105.0 * y3 * z * r0[7] + 3.0 * y * z7 * r0[9] + 63.0 * y * z5 * r0[8] +
    315.0 * y * z3 * r0[7] + 315.0 * y * z * r0[6];
  r[1][2][7] =
    x * y2 * z7 * r0[10] + 21.0 * x * y2 * z5 * r0[9] +
    105.0 * x * y2 * z3 * r0[8] + 105.0 * x * y2 * z * r0[7] +
    x * z7 * r0[9] + 21.0 * x * z5 * r0[8] + 105.0 * x * z3 * r0[7] +
    105.0 * x * z * r0[6];
  r[2][1][7] =
    x2 * y * z7 * r0[10] + 21.0 * x2 * y * z5 * r0[9] +
    105.0 * x2 * y * z3 * r0[8] + 105.0 * x2 * y * z * r0[7] +
    y * z7 * r0[9] + 21.0 * y * z5 * r0[8] + 105.0 * y * z3 * r0[7] +
    105.0 * y * z * r0[6];
  r[3][0][7] =
    x3 * z7 * r0[10] + 21.0 * x3 * z5 * r0[9] + 105.0 * x3 * z3 * r0[8] +
    105.0 * x3 * z * r0[7] + 3.0 * x * z7 * r0[9] + 63.0 * x * z5 * r0[8] +
    315.0 * x * z3 * r0[7] + 315.0 * x * z * r0[6];
  r[0][4][6] =
    y4 * z6 * r0[10] + 15.0 * y4 * z4 * r0[9] + 45.0 * y4 * z2 * r0[8] +
    15.0 * y4 * r0[7] + 6.0 * y2 * z6 * r0[9] + 90.0 * y2 * z4 * r0[8] +
    270.0 * y2 * z2 * r0[7] + 90.0 * y2 * r0[6] + 3.0 * z6 * r0[8] +
    45.0 * z4 * r0[7] + 135.0 * z2 * r0[6] + 45.0 * r0[5];
  r[1][3][6] =
    x * y3 * z6 * r0[10] + 15.0 * x * y3 * z4 * r0[9] +
    45.0 * x * y3 * z2 * r0[8] + 15.0 * x * y3 * r0[7] +
    3.0 * x * y * z6 * r0[9] + 45.0 * x * y * z4 * r0[8] +
    135.0 * x * y * z2 * r0[7] + 45.0 * x * y * r0[6];
  r[2][2][6] =
    x2 * y2 * z6 * r0[10] + 15.0 * x2 * y2 * z4 * r0[9] +
    45.0 * x2 * y2 * z2 * r0[8] + 15.0 * x2 * y2 * r0[7] + x2 * z6 * r0[9] +
    15.0 * x2 * z4 * r0[8] + 45.0 * x2 * z2 * r0[7] + 15.0 * x2 * r0[6] +
    y2 * z6 * r0[9] + 15.0 * y2 * z4 * r0[8] + 45.0 * y2 * z2 * r0[7] +
    15.0 * y2 * r0[6] + z6 * r0[8] + 15.0 * z4 * r0[7] + 45.0 * z2 * r0[6] +
    15.0 * r0[5];
  r[3][1][6] =
    x3 * y * z6 * r0[10] + 15.0 * x3 * y * z4 * r0[9] +
    45.0 * x3 * y * z2 * r0[8] + 15.0 * x3 * y * r0[7] +
    3.0 * x * y * z6 * r0[9] + 45.0 * x * y * z4 * r0[8] +
    135.0 * x * y * z2 * r0[7] + 45.0 * x * y * r0[6];
  r[4][0][6] =
    x4 * z6 * r0[10] + 15.0 * x4 * z4 * r0[9] + 45.0 * x4 * z2 * r0[8] +
    15.0 * x4 * r0[7] + 6.0 * x2 * z6 * r0[9] + 90.0 * x2 * z4 * r0[8] +
    270.0 * x2 * z2 * r0[7] + 90.0 * x2 * r0[6] + 3.0 * z6 * r0[8] +
    45.0 * z4 * r0[7] + 135.0 * z2 * r0[6] + 45.0 * r0[5];
  r[0][5][5] =
    y5 * z5 * r0[10] + 10.0 * y5 * z3 * r0[9] + 15.0 * y5 * z * r0[8] +
    10.0 * y3 * z5 * r0[9] + 100.0 * y3 * z3 * r0[8] +
    150.0 * y3 * z * r0[7] + 15.0 * y * z5 * r0[8] + 150.0 * y * z3 * r0[7] +
    225.0 * y * z * r0[6];
  r[1][4][5] =
    x * y4 * z5 * r0[10] + 10.0 * x * y4 * z3 * r0[9] +
    15.0 * x * y4 * z * r0[8] + 6.0 * x * y2 * z5 * r0[9] +
    60.0 * x * y2 * z3 * r0[8] + 90.0 * x * y2 * z * r0[7] +
    3.0 * x * z5 * r0[8] + 30.0 * x * z3 * r0[7] + 45.0 * x * z * r0[6];
  r[2][3][5] =
    x2 * y3 * z5 * r0[10] + 10.0 * x2 * y3 * z3 * r0[9] +
    15.0 * x2 * y3 * z * r0[8] + 3.0 * x2 * y * z5 * r0[9] +
    30.0 * x2 * y * z3 * r0[8] + 45.0 * x2 * y * z * r0[7] + y3 * z5 * r0[9] +
    10.0 * y3 * z3 * r0[8] + 15.0 * y3 * z * r0[7] + 3.0 * y * z5 * r0[8] +
    30.0 * y * z3 * r0[7] + 45.0 * y * z * r0[6];
  r[3][2][5] =
    x3 * y2 * z5 * r0[10] + 10.0 * x3 * y2 * z3 * r0[9] +
    15.0 * x3 * y2 * z * r0[8] + x3 * z5 * r0[9] + 10.0 * x3 * z3 * r0[8] +
    15.0 * x3 * z * r0[7] + 3.0 * x * y2 * z5 * r0[9] +
    30.0 * x * y2 * z3 * r0[8] + 45.0 * x * y2 * z * r0[7] +
    3.0 * x * z5 * r0[8] + 30.0 * x * z3 * r0[7] + 45.0 * x * z * r0[6];
  r[4][1][5] =
    x4 * y * z5 * r0[10] + 10.0 * x4 * y * z3 * r0[9] +
    15.0 * x4 * y * z * r0[8] + 6.0 * x2 * y * z5 * r0[9] +
    60.0 * x2 * y * z3 * r0[8] + 90.0 * x2 * y * z * r0[7] +
    3.0 * y * z5 * r0[8] + 30.0 * y * z3 * r0[7] + 45.0 * y * z * r0[6];
  r[5][0][5] =
    x5 * z5 * r0[10] + 10.0 * x5 * z3 * r0[9] + 15.0 * x5 * z * r0[8] +
    10.0 * x3 * z5 * r0[9] + 100.0 * x3 * z3 * r0[8] +
    150.0 * x3 * z * r0[7] + 15.0 * x * z5 * r0[8] + 150.0 * x * z3 * r0[7] +
    225.0 * x * z * r0[6];
  r[0][6][4] =
    y6 * z4 * r0[10] + 6.0 * y6 * z2 * r0[9] + 3.0 * y6 * r0[8] +
    15.0 * y4 * z4 * r0[9] + 90.0 * y4 * z2 * r0[8] + 45.0 * y4 * r0[7] +
    45.0 * y2 * z4 * r0[8] + 270.0 * y2 * z2 * r0[7] + 135.0 * y2 * r0[6] +
    15.0 * z4 * r0[7] + 90.0 * z2 * r0[6] + 45.0 * r0[5];
  r[1][5][4] =
    x * y5 * z4 * r0[10] + 6.0 * x * y5 * z2 * r0[9] + 3.0 * x * y5 * r0[8] +
    10.0 * x * y3 * z4 * r0[9] + 60.0 * x * y3 * z2 * r0[8] +
    30.0 * x * y3 * r0[7] + 15.0 * x * y * z4 * r0[8] +
    90.0 * x * y * z2 * r0[7] + 45.0 * x * y * r0[6];
  r[2][4][4] =
    x2 * y4 * z4 * r0[10] + 6.0 * x2 * y4 * z2 * r0[9] +
    3.0 * x2 * y4 * r0[8] + 6.0 * x2 * y2 * z4 * r0[9] +
    36.0 * x2 * y2 * z2 * r0[8] + 18.0 * x2 * y2 * r0[7] +
    3.0 * x2 * z4 * r0[8] + 18.0 * x2 * z2 * r0[7] + 9.0 * x2 * r0[6] +
    y4 * z4 * r0[9] + 6.0 * y4 * z2 * r0[8] + 3.0 * y4 * r0[7] +
    6.0 * y2 * z4 * r0[8] + 36.0 * y2 * z2 * r0[7] + 18.0 * y2 * r0[6] +
    3.0 * z4 * r0[7] + 18.0 * z2 * r0[6] + 9.0 * r0[5];
  r[3][3][4] =
    x3 * y3 * z4 * r0[10] + 6.0 * x3 * y3 * z2 * r0[9] +
    3.0 * x3 * y3 * r0[8] + 3.0 * x3 * y * z4 * r0[9] +
    18.0 * x3 * y * z2 * r0[8] + 9.0 * x3 * y * r0[7] +
    3.0 * x * y3 * z4 * r0[9] + 18.0 * x * y3 * z2 * r0[8] +
    9.0 * x * y3 * r0[7] + 9.0 * x * y * z4 * r0[8] +
    54.0 * x * y * z2 * r0[7] + 27.0 * x * y * r0[6];
  r[4][2][4] =
    x4 * y2 * z4 * r0[10] + 6.0 * x4 * y2 * z2 * r0[9] +
    3.0 * x4 * y2 * r0[8] + x4 * z4 * r0[9] + 6.0 * x4 * z2 * r0[8] +
    3.0 * x4 * r0[7] + 6.0 * x2 * y2 * z4 * r0[9] +
    36.0 * x2 * y2 * z2 * r0[8] + 18.0 * x2 * y2 * r0[7] +
    6.0 * x2 * z4 * r0[8] + 36.0 * x2 * z2 * r0[7] + 18.0 * x2 * r0[6] +
    3.0 * y2 * z4 * r0[8] + 18.0 * y2 * z2 * r0[7] + 9.0 * y2 * r0[6] +
    3.0 * z4 * r0[7] + 18.0 * z2 * r0[6] + 9.0 * r0[5];
  r[5][1][4] =
    x5 * y * z4 * r0[10] + 6.0 * x5 * y * z2 * r0[9] + 3.0 * x5 * y * r0[8] +
    10.0 * x3 * y * z4 * r0[9] + 60.0 * x3 * y * z2 * r0[8] +
    30.0 * x3 * y * r0[7] + 15.0 * x * y * z4 * r0[8] +
    90.0 * x * y * z2 * r0[7] + 45.0 * x * y * r0[6];
  r[6][0][4] =
    x6 * z4 * r0[10] + 6.0 * x6 * z2 * r0[9] + 3.0 * x6 * r0[8] +
    15.0 * x4 * z4 * r0[9] + 90.0 * x4 * z2 * r0[8] + 45.0 * x4 * r0[7] +
    45.0 * x2 * z4 * r0[8] + 270.0 * x2 * z2 * r0[7] + 135.0 * x2 * r0[6] +
    15.0 * z4 * r0[7] + 90.0 * z2 * r0[6] + 45.0 * r0[5];
  r[0][7][3] =
    y7 * z3 * r0[10] + 3.0 * y7 * z * r0[9] + 21.0 * y5 * z3 * r0[9] +
    63.0 * y5 * z * r0[8] + 105.0 * y3 * z3 * r0[8] + 315.0 * y3 * z * r0[7] +
    105.0 * y * z3 * r0[7] + 315.0 * y * z * r0[6];
  r[1][6][3] =
    x * y6 * z3 * r0[10] + 3.0 * x * y6 * z * r0[9] +
    15.0 * x * y4 * z3 * r0[9] + 45.0 * x * y4 * z * r0[8] +
    45.0 * x * y2 * z3 * r0[8] + 135.0 * x * y2 * z * r0[7] +
    15.0 * x * z3 * r0[7] + 45.0 * x * z * r0[6];
  r[2][5][3] =
    x2 * y5 * z3 * r0[10] + 3.0 * x2 * y5 * z * r0[9] +
    10.0 * x2 * y3 * z3 * r0[9] + 30.0 * x2 * y3 * z * r0[8] +
    15.0 * x2 * y * z3 * r0[8] + 45.0 * x2 * y * z * r0[7] + y5 * z3 * r0[9] +
    3.0 * y5 * z * r0[8] + 10.0 * y3 * z3 * r0[8] + 30.0 * y3 * z * r0[7] +
    15.0 * y * z3 * r0[7] + 45.0 * y * z * r0[6];
  r[3][4][3] =
    x3 * y4 * z3 * r0[10] + 3.0 * x3 * y4 * z * r0[9] +
    6.0 * x3 * y2 * z3 * r0[9] + 18.0 * x3 * y2 * z * r0[8] +
    3.0 * x3 * z3 * r0[8] + 9.0 * x3 * z * r0[7] + 3.0 * x * y4 * z3 * r0[9] +
    9.0 * x * y4 * z * r0[8] + 18.0 * x * y2 * z3 * r0[8] +
    54.0 * x * y2 * z * r0[7] + 9.0 * x * z3 * r0[7] + 27.0 * x * z * r0[6];
  r[4][3][3] =
    x4 * y3 * z3 * r0[10] + 3.0 * x4 * y3 * z * r0[9] +
    3.0 * x4 * y * z3 * r0[9] + 9.0 * x4 * y * z * r0[8] +
    6.0 * x2 * y3 * z3 * r0[9] + 18.0 * x2 * y3 * z * r0[8] +
    18.0 * x2 * y * z3 * r0[8] + 54.0 * x2 * y * z * r0[7] +
    3.0 * y3 * z3 * r0[8] + 9.0 * y3 * z * r0[7] + 9.0 * y * z3 * r0[7] +
    27.0 * y * z * r0[6];
  r[5][2][3] =
    x5 * y2 * z3 * r0[10] + 3.0 * x5 * y2 * z * r0[9] + x5 * z3 * r0[9] +
    3.0 * x5 * z * r0[8] + 10.0 * x3 * y2 * z3 * r0[9] +
    30.0 * x3 * y2 * z * r0[8] + 10.0 * x3 * z3 * r0[8] +
    30.0 * x3 * z * r0[7] + 15.0 * x * y2 * z3 * r0[8] +
    45.0 * x * y2 * z * r0[7] + 15.0 * x * z3 * r0[7] + 45.0 * x * z * r0[6];
  r[6][1][3] =
    x6 * y * z3 * r0[10] + 3.0 * x6 * y * z * r0[9] +
    15.0 * x4 * y * z3 * r0[9] + 45.0 * x4 * y * z * r0[8] +
    45.0 * x2 * y * z3 * r0[8] + 135.0 * x2 * y * z * r0[7] +
    15.0 * y * z3 * r0[7] + 45.0 * y * z * r0[6];
  r[7][0][3] =
    x7 * z3 * r0[10] + 3.0 * x7 * z * r0[9] + 21.0 * x5 * z3 * r0[9] +
    63.0 * x5 * z * r0[8] + 105.0 * x3 * z3 * r0[8] + 315.0 * x3 * z * r0[7] +
    105.0 * x * z3 * r0[7] + 315.0 * x * z * r0[6];
  r[0][8][2] =
    y8 * z2 * r0[10] + y8 * r0[9] + 28.0 * y6 * z2 * r0[9] +
    28.0 * y6 * r0[8] + 210.0 * y4 * z2 * r0[8] + 210.0 * y4 * r0[7] +
    420.0 * y2 * z2 * r0[7] + 420.0 * y2 * r0[6] + 105.0 * z2 * r0[6] +
    105.0 * r0[5];
  r[1][7][2] =
    x * y7 * z2 * r0[10] + x * y7 * r0[9] + 21.0 * x * y5 * z2 * r0[9] +
    21.0 * x * y5 * r0[8] + 105.0 * x * y3 * z2 * r0[8] +
    105.0 * x * y3 * r0[7] + 105.0 * x * y * z2 * r0[7] +
    105.0 * x * y * r0[6];
  r[2][6][2] =
    x2 * y6 * z2 * r0[10] + x2 * y6 * r0[9] + 15.0 * x2 * y4 * z2 * r0[9] +
    15.0 * x2 * y4 * r0[8] + 45.0 * x2 * y2 * z2 * r0[8] +
    45.0 * x2 * y2 * r0[7] + 15.0 * x2 * z2 * r0[7] + 15.0 * x2 * r0[6] +
    y6 * z2 * r0[9] + y6 * r0[8] + 15.0 * y4 * z2 * r0[8] +
    15.0 * y4 * r0[7] + 45.0 * y2 * z2 * r0[7] + 45.0 * y2 * r0[6] +
    15.0 * z2 * r0[6] + 15.0 * r0[5];
  r[3][5][2] =
    x3 * y5 * z2 * r0[10] + x3 * y5 * r0[9] + 10.0 * x3 * y3 * z2 * r0[9] +
    10.0 * x3 * y3 * r0[8] + 15.0 * x3 * y * z2 * r0[8] +
    15.0 * x3 * y * r0[7] + 3.0 * x * y5 * z2 * r0[9] + 3.0 * x * y5 * r0[8] +
    30.0 * x * y3 * z2 * r0[8] + 30.0 * x * y3 * r0[7] +
    45.0 * x * y * z2 * r0[7] + 45.0 * x * y * r0[6];
  r[4][4][2] =
    x4 * y4 * z2 * r0[10] + x4 * y4 * r0[9] + 6.0 * x4 * y2 * z2 * r0[9] +
    6.0 * x4 * y2 * r0[8] + 3.0 * x4 * z2 * r0[8] + 3.0 * x4 * r0[7] +
    6.0 * x2 * y4 * z2 * r0[9] + 6.0 * x2 * y4 * r0[8] +
    36.0 * x2 * y2 * z2 * r0[8] + 36.0 * x2 * y2 * r0[7] +
    18.0 * x2 * z2 * r0[7] + 18.0 * x2 * r0[6] + 3.0 * y4 * z2 * r0[8] +
    3.0 * y4 * r0[7] + 18.0 * y2 * z2 * r0[7] + 18.0 * y2 * r0[6] +
    9.0 * z2 * r0[6] + 9.0 * r0[5];
  r[5][3][2] =
    x5 * y3 * z2 * r0[10] + x5 * y3 * r0[9] + 3.0 * x5 * y * z2 * r0[9] +
    3.0 * x5 * y * r0[8] + 10.0 * x3 * y3 * z2 * r0[9] +
    10.0 * x3 * y3 * r0[8] + 30.0 * x3 * y * z2 * r0[8] +
    30.0 * x3 * y * r0[7] + 15.0 * x * y3 * z2 * r0[8] +
    15.0 * x * y3 * r0[7] + 45.0 * x * y * z2 * r0[7] + 45.0 * x * y * r0[6];
  r[6][2][2] =
    x6 * y2 * z2 * r0[10] + x6 * y2 * r0[9] + x6 * z2 * r0[9] + x6 * r0[8] +
    15.0 * x4 * y2 * z2 * r0[9] + 15.0 * x4 * y2 * r0[8] +
    15.0 * x4 * z2 * r0[8] + 15.0 * x4 * r0[7] + 45.0 * x2 * y2 * z2 * r0[8] +
    45.0 * x2 * y2 * r0[7] + 45.0 * x2 * z2 * r0[7] + 45.0 * x2 * r0[6] +
    15.0 * y2 * z2 * r0[7] + 15.0 * y2 * r0[6] + 15.0 * z2 * r0[6] +
    15.0 * r0[5];
  r[7][1][2] =
    x7 * y * z2 * r0[10] + x7 * y * r0[9] + 21.0 * x5 * y * z2 * r0[9] +
    21.0 * x5 * y * r0[8] + 105.0 * x3 * y * z2 * r0[8] +
    105.0 * x3 * y * r0[7] + 105.0 * x * y * z2 * r0[7] +
    105.0 * x * y * r0[6];
  r[8][0][2] =
    x8 * z2 * r0[10] + x8 * r0[9] + 28.0 * x6 * z2 * r0[9] +
    28.0 * x6 * r0[8] + 210.0 * x4 * z2 * r0[8] + 210.0 * x4 * r0[7] +
    420.0 * x2 * z2 * r0[7] + 420.0 * x2 * r0[6] + 105.0 * z2 * r0[6] +
    105.0 * r0[5];
  r[0][9][1] =
    y9 * z * r0[10] + 36.0 * y7 * z * r0[9] + 378.0 * y5 * z * r0[8] +
    1260.0 * y3 * z * r0[7] + 945.0 * y * z * r0[6];
  r[1][8][1] =
    x * y8 * z * r0[10] + 28.0 * x * y6 * z * r0[9] +
    210.0 * x * y4 * z * r0[8] + 420.0 * x * y2 * z * r0[7] +
    105.0 * x * z * r0[6];
  r[2][7][1] =
    x2 * y7 * z * r0[10] + 21.0 * x2 * y5 * z * r0[9] +
    105.0 * x2 * y3 * z * r0[8] + 105.0 * x2 * y * z * r0[7] +
    y7 * z * r0[9] + 21.0 * y5 * z * r0[8] + 105.0 * y3 * z * r0[7] +
    105.0 * y * z * r0[6];
  r[3][6][1] =
    x3 * y6 * z * r0[10] + 15.0 * x3 * y4 * z * r0[9] +
    45.0 * x3 * y2 * z * r0[8] + 15.0 * x3 * z * r0[7] +
    3.0 * x * y6 * z * r0[9] + 45.0 * x * y4 * z * r0[8] +
    135.0 * x * y2 * z * r0[7] + 45.0 * x * z * r0[6];
  r[4][5][1] =
    x4 * y5 * z * r0[10] + 10.0 * x4 * y3 * z * r0[9] +
    15.0 * x4 * y * z * r0[8] + 6.0 * x2 * y5 * z * r0[9] +
    60.0 * x2 * y3 * z * r0[8] + 90.0 * x2 * y * z * r0[7] +
    3.0 * y5 * z * r0[8] + 30.0 * y3 * z * r0[7] + 45.0 * y * z * r0[6];
  r[5][4][1] =
    x5 * y4 * z * r0[10] + 6.0 * x5 * y2 * z * r0[9] + 3.0 * x5 * z * r0[8] +
    10.0 * x3 * y4 * z * r0[9] + 60.0 * x3 * y2 * z * r0[8] +
    30.0 * x3 * z * r0[7] + 15.0 * x * y4 * z * r0[8] +
    90.0 * x * y2 * z * r0[7] + 45.0 * x * z * r0[6];
  r[6][3][1] =
    x6 * y3 * z * r0[10] + 3.0 * x6 * y * z * r0[9] +
    15.0 * x4 * y3 * z * r0[9] + 45.0 * x4 * y * z * r0[8] +
    45.0 * x2 * y3 * z * r0[8] + 135.0 * x2 * y * z * r0[7] +
    15.0 * y3 * z * r0[7] + 45.0 * y * z * r0[6];
  r[7][2][1] =
    x7 * y2 * z * r0[10] + x7 * z * r0[9] + 21.0 * x5 * y2 * z * r0[9] +
    21.0 * x5 * z * r0[8] + 105.0 * x3 * y2 * z * r0[8] +
    105.0 * x3 * z * r0[7] + 105.0 * x * y2 * z * r0[7] +
    105.0 * x * z * r0[6];
  r[8][1][1] =
    x8 * y * z * r0[10] + 28.0 * x6 * y * z * r0[9] +
    210.0 * x4 * y * z * r0[8] + 420.0 * x2 * y * z * r0[7] +
    105.0 * y * z * r0[6];
  r[9][0][1] =
    x9 * z * r0[10] + 36.0 * x7 * z * r0[9] + 378.0 * x5 * z * r0[8] +
    1260.0 * x3 * z * r0[7] + 945.0 * x * z * r0[6];
  r[0][10][0] =
    y10 * r0[10] + 45.0 * y8 * r0[9] + 630.0 * y6 * r0[8] +
    3150.0 * y4 * r0[7] + 4725.0 * y2 * r0[6] + 945.0 * r0[5];
  r[1][9][0] =
    x * y9 * r0[10] + 36.0 * x * y7 * r0[9] + 378.0 * x * y5 * r0[8] +
    1260.0 * x * y3 * r0[7] + 945.0 * x * y * r0[6];
  r[2][8][0] =
    x2 * y8 * r0[10] + 28.0 * x2 * y6 * r0[9] + 210.0 * x2 * y4 * r0[8] +
    420.0 * x2 * y2 * r0[7] + 105.0 * x2 * r0[6] + y8 * r0[9] +
    28.0 * y6 * r0[8] + 210.0 * y4 * r0[7] + 420.0 * y2 * r0[6] +
    105.0 * r0[5];
  r[3][7][0] =
    x3 * y7 * r0[10] + 21.0 * x3 * y5 * r0[9] + 105.0 * x3 * y3 * r0[8] +
    105.0 * x3 * y * r0[7] + 3.0 * x * y7 * r0[9] + 63.0 * x * y5 * r0[8] +
    315.0 * x * y3 * r0[7] + 315.0 * x * y * r0[6];
  r[4][6][0] =
    x4 * y6 * r0[10] + 15.0 * x4 * y4 * r0[9] + 45.0 * x4 * y2 * r0[8] +
    15.0 * x4 * r0[7] + 6.0 * x2 * y6 * r0[9] + 90.0 * x2 * y4 * r0[8] +
    270.0 * x2 * y2 * r0[7] + 90.0 * x2 * r0[6] + 3.0 * y6 * r0[8] +
    45.0 * y4 * r0[7] + 135.0 * y2 * r0[6] + 45.0 * r0[5];
  r[5][5][0] =
    x5 * y5 * r0[10] + 10.0 * x5 * y3 * r0[9] + 15.0 * x5 * y * r0[8] +
    10.0 * x3 * y5 * r0[9] + 100.0 * x3 * y3 * r0[8] +
    150.0 * x3 * y * r0[7] + 15.0 * x * y5 * r0[8] + 150.0 * x * y3 * r0[7] +
    225.0 * x * y * r0[6];
  r[6][4][0] =
    x6 * y4 * r0[10] + 6.0 * x6 * y2 * r0[9] + 3.0 * x6 * r0[8] +
    15.0 * x4 * y4 * r0[9] + 90.0 * x4 * y2 * r0[8] + 45.0 * x4 * r0[7] +
    45.0 * x2 * y4 * r0[8] + 270.0 * x2 * y2 * r0[7] + 135.0 * x2 * r0[6] +
    15.0 * y4 * r0[7] + 90.0 * y2 * r0[6] + 45.0 * r0[5];
  r[7][3][0] =
    x7 * y3 * r0[10] + 3.0 * x7 * y * r0[9] + 21.0 * x5 * y3 * r0[9] +
    63.0 * x5 * y * r0[8] + 105.0 * x3 * y3 * r0[8] + 315.0 * x3 * y * r0[7] +
    105.0 * x * y3 * r0[7] + 315.0 * x * y * r0[6];
  r[8][2][0] =
    x8 * y2 * r0[10] + x8 * r0[9] + 28.0 * x6 * y2 * r0[9] +
    28.0 * x6 * r0[8] + 210.0 * x4 * y2 * r0[8] + 210.0 * x4 * r0[7] +
    420.0 * x2 * y2 * r0[7] + 420.0 * x2 * r0[6] + 105.0 * y2 * r0[6] +
    105.0 * r0[5];
  r[9][1][0] =
    x9 * y * r0[10] + 36.0 * x7 * y * r0[9] + 378.0 * x5 * y * r0[8] +
    1260.0 * x3 * y * r0[7] + 945.0 * x * y * r0[6];
  r[10][0][0] =
    x10 * r0[10] + 45.0 * x8 * r0[9] + 630.0 * x6 * r0[8] +
    3150.0 * x4 * r0[7] + 4725.0 * x2 * r0[6] + 945.0 * r0[5];
  if (ltot == 10)
    {
      return;
    }
  x11 = x10 * x;
  y11 = y10 * y;
  z11 = z10 * z;
  r[0][0][11] =
    z11 * r0[11] + 55.0 * z9 * r0[10] + 990.0 * z7 * r0[9] +
    6930.0 * z5 * r0[8] + 17325.0 * z3 * r0[7] + 10395.0 * z * r0[6];
  r[0][1][10] =
    y * z10 * r0[11] + 45.0 * y * z8 * r0[10] + 630.0 * y * z6 * r0[9] +
    3150.0 * y * z4 * r0[8] + 4725.0 * y * z2 * r0[7] + 945.0 * y * r0[6];
  r[1][0][10] =
    x * z10 * r0[11] + 45.0 * x * z8 * r0[10] + 630.0 * x * z6 * r0[9] +
    3150.0 * x * z4 * r0[8] + 4725.0 * x * z2 * r0[7] + 945.0 * x * r0[6];
  r[0][2][9] =
    y2 * z9 * r0[11] + 36.0 * y2 * z7 * r0[10] + 378.0 * y2 * z5 * r0[9] +
    1260.0 * y2 * z3 * r0[8] + 945.0 * y2 * z * r0[7] + z9 * r0[10] +
    36.0 * z7 * r0[9] + 378.0 * z5 * r0[8] + 1260.0 * z3 * r0[7] +
    945.0 * z * r0[6];
  r[1][1][9] =
    x * y * z9 * r0[11] + 36.0 * x * y * z7 * r0[10] +
    378.0 * x * y * z5 * r0[9] + 1260.0 * x * y * z3 * r0[8] +
    945.0 * x * y * z * r0[7];
  r[2][0][9] =
    x2 * z9 * r0[11] + 36.0 * x2 * z7 * r0[10] + 378.0 * x2 * z5 * r0[9] +
    1260.0 * x2 * z3 * r0[8] + 945.0 * x2 * z * r0[7] + z9 * r0[10] +
    36.0 * z7 * r0[9] + 378.0 * z5 * r0[8] + 1260.0 * z3 * r0[7] +
    945.0 * z * r0[6];
  r[0][3][8] =
    y3 * z8 * r0[11] + 28.0 * y3 * z6 * r0[10] + 210.0 * y3 * z4 * r0[9] +
    420.0 * y3 * z2 * r0[8] + 105.0 * y3 * r0[7] + 3.0 * y * z8 * r0[10] +
    84.0 * y * z6 * r0[9] + 630.0 * y * z4 * r0[8] + 1260.0 * y * z2 * r0[7] +
    315.0 * y * r0[6];
  r[1][2][8] =
    x * y2 * z8 * r0[11] + 28.0 * x * y2 * z6 * r0[10] +
    210.0 * x * y2 * z4 * r0[9] + 420.0 * x * y2 * z2 * r0[8] +
    105.0 * x * y2 * r0[7] + x * z8 * r0[10] + 28.0 * x * z6 * r0[9] +
    210.0 * x * z4 * r0[8] + 420.0 * x * z2 * r0[7] + 105.0 * x * r0[6];
  r[2][1][8] =
    x2 * y * z8 * r0[11] + 28.0 * x2 * y * z6 * r0[10] +
    210.0 * x2 * y * z4 * r0[9] + 420.0 * x2 * y * z2 * r0[8] +
    105.0 * x2 * y * r0[7] + y * z8 * r0[10] + 28.0 * y * z6 * r0[9] +
    210.0 * y * z4 * r0[8] + 420.0 * y * z2 * r0[7] + 105.0 * y * r0[6];
  r[3][0][8] =
    x3 * z8 * r0[11] + 28.0 * x3 * z6 * r0[10] + 210.0 * x3 * z4 * r0[9] +
    420.0 * x3 * z2 * r0[8] + 105.0 * x3 * r0[7] + 3.0 * x * z8 * r0[10] +
    84.0 * x * z6 * r0[9] + 630.0 * x * z4 * r0[8] + 1260.0 * x * z2 * r0[7] +
    315.0 * x * r0[6];
  r[0][4][7] =
    y4 * z7 * r0[11] + 21.0 * y4 * z5 * r0[10] + 105.0 * y4 * z3 * r0[9] +
    105.0 * y4 * z * r0[8] + 6.0 * y2 * z7 * r0[10] +
    126.0 * y2 * z5 * r0[9] + 630.0 * y2 * z3 * r0[8] +
    630.0 * y2 * z * r0[7] + 3.0 * z7 * r0[9] + 63.0 * z5 * r0[8] +
    315.0 * z3 * r0[7] + 315.0 * z * r0[6];
  r[1][3][7] =
    x * y3 * z7 * r0[11] + 21.0 * x * y3 * z5 * r0[10] +
    105.0 * x * y3 * z3 * r0[9] + 105.0 * x * y3 * z * r0[8] +
    3.0 * x * y * z7 * r0[10] + 63.0 * x * y * z5 * r0[9] +
    315.0 * x * y * z3 * r0[8] + 315.0 * x * y * z * r0[7];
  r[2][2][7] =
    x2 * y2 * z7 * r0[11] + 21.0 * x2 * y2 * z5 * r0[10] +
    105.0 * x2 * y2 * z3 * r0[9] + 105.0 * x2 * y2 * z * r0[8] +
    x2 * z7 * r0[10] + 21.0 * x2 * z5 * r0[9] + 105.0 * x2 * z3 * r0[8] +
    105.0 * x2 * z * r0[7] + y2 * z7 * r0[10] + 21.0 * y2 * z5 * r0[9] +
    105.0 * y2 * z3 * r0[8] + 105.0 * y2 * z * r0[7] + z7 * r0[9] +
    21.0 * z5 * r0[8] + 105.0 * z3 * r0[7] + 105.0 * z * r0[6];
  r[3][1][7] =
    x3 * y * z7 * r0[11] + 21.0 * x3 * y * z5 * r0[10] +
    105.0 * x3 * y * z3 * r0[9] + 105.0 * x3 * y * z * r0[8] +
    3.0 * x * y * z7 * r0[10] + 63.0 * x * y * z5 * r0[9] +
    315.0 * x * y * z3 * r0[8] + 315.0 * x * y * z * r0[7];
  r[4][0][7] =
    x4 * z7 * r0[11] + 21.0 * x4 * z5 * r0[10] + 105.0 * x4 * z3 * r0[9] +
    105.0 * x4 * z * r0[8] + 6.0 * x2 * z7 * r0[10] +
    126.0 * x2 * z5 * r0[9] + 630.0 * x2 * z3 * r0[8] +
    630.0 * x2 * z * r0[7] + 3.0 * z7 * r0[9] + 63.0 * z5 * r0[8] +
    315.0 * z3 * r0[7] + 315.0 * z * r0[6];
  r[0][5][6] =
    y5 * z6 * r0[11] + 15.0 * y5 * z4 * r0[10] + 45.0 * y5 * z2 * r0[9] +
    15.0 * y5 * r0[8] + 10.0 * y3 * z6 * r0[10] + 150.0 * y3 * z4 * r0[9] +
    450.0 * y3 * z2 * r0[8] + 150.0 * y3 * r0[7] + 15.0 * y * z6 * r0[9] +
    225.0 * y * z4 * r0[8] + 675.0 * y * z2 * r0[7] + 225.0 * y * r0[6];
  r[1][4][6] =
    x * y4 * z6 * r0[11] + 15.0 * x * y4 * z4 * r0[10] +
    45.0 * x * y4 * z2 * r0[9] + 15.0 * x * y4 * r0[8] +
    6.0 * x * y2 * z6 * r0[10] + 90.0 * x * y2 * z4 * r0[9] +
    270.0 * x * y2 * z2 * r0[8] + 90.0 * x * y2 * r0[7] +
    3.0 * x * z6 * r0[9] + 45.0 * x * z4 * r0[8] + 135.0 * x * z2 * r0[7] +
    45.0 * x * r0[6];
  r[2][3][6] =
    x2 * y3 * z6 * r0[11] + 15.0 * x2 * y3 * z4 * r0[10] +
    45.0 * x2 * y3 * z2 * r0[9] + 15.0 * x2 * y3 * r0[8] +
    3.0 * x2 * y * z6 * r0[10] + 45.0 * x2 * y * z4 * r0[9] +
    135.0 * x2 * y * z2 * r0[8] + 45.0 * x2 * y * r0[7] + y3 * z6 * r0[10] +
    15.0 * y3 * z4 * r0[9] + 45.0 * y3 * z2 * r0[8] + 15.0 * y3 * r0[7] +
    3.0 * y * z6 * r0[9] + 45.0 * y * z4 * r0[8] + 135.0 * y * z2 * r0[7] +
    45.0 * y * r0[6];
  r[3][2][6] =
    x3 * y2 * z6 * r0[11] + 15.0 * x3 * y2 * z4 * r0[10] +
    45.0 * x3 * y2 * z2 * r0[9] + 15.0 * x3 * y2 * r0[8] + x3 * z6 * r0[10] +
    15.0 * x3 * z4 * r0[9] + 45.0 * x3 * z2 * r0[8] + 15.0 * x3 * r0[7] +
    3.0 * x * y2 * z6 * r0[10] + 45.0 * x * y2 * z4 * r0[9] +
    135.0 * x * y2 * z2 * r0[8] + 45.0 * x * y2 * r0[7] +
    3.0 * x * z6 * r0[9] + 45.0 * x * z4 * r0[8] + 135.0 * x * z2 * r0[7] +
    45.0 * x * r0[6];
  r[4][1][6] =
    x4 * y * z6 * r0[11] + 15.0 * x4 * y * z4 * r0[10] +
    45.0 * x4 * y * z2 * r0[9] + 15.0 * x4 * y * r0[8] +
    6.0 * x2 * y * z6 * r0[10] + 90.0 * x2 * y * z4 * r0[9] +
    270.0 * x2 * y * z2 * r0[8] + 90.0 * x2 * y * r0[7] +
    3.0 * y * z6 * r0[9] + 45.0 * y * z4 * r0[8] + 135.0 * y * z2 * r0[7] +
    45.0 * y * r0[6];
  r[5][0][6] =
    x5 * z6 * r0[11] + 15.0 * x5 * z4 * r0[10] + 45.0 * x5 * z2 * r0[9] +
    15.0 * x5 * r0[8] + 10.0 * x3 * z6 * r0[10] + 150.0 * x3 * z4 * r0[9] +
    450.0 * x3 * z2 * r0[8] + 150.0 * x3 * r0[7] + 15.0 * x * z6 * r0[9] +
    225.0 * x * z4 * r0[8] + 675.0 * x * z2 * r0[7] + 225.0 * x * r0[6];
  r[0][6][5] =
    y6 * z5 * r0[11] + 10.0 * y6 * z3 * r0[10] + 15.0 * y6 * z * r0[9] +
    15.0 * y4 * z5 * r0[10] + 150.0 * y4 * z3 * r0[9] +
    225.0 * y4 * z * r0[8] + 45.0 * y2 * z5 * r0[9] +
    450.0 * y2 * z3 * r0[8] + 675.0 * y2 * z * r0[7] + 15.0 * z5 * r0[8] +
    150.0 * z3 * r0[7] + 225.0 * z * r0[6];
  r[1][5][5] =
    x * y5 * z5 * r0[11] + 10.0 * x * y5 * z3 * r0[10] +
    15.0 * x * y5 * z * r0[9] + 10.0 * x * y3 * z5 * r0[10] +
    100.0 * x * y3 * z3 * r0[9] + 150.0 * x * y3 * z * r0[8] +
    15.0 * x * y * z5 * r0[9] + 150.0 * x * y * z3 * r0[8] +
    225.0 * x * y * z * r0[7];
  r[2][4][5] =
    x2 * y4 * z5 * r0[11] + 10.0 * x2 * y4 * z3 * r0[10] +
    15.0 * x2 * y4 * z * r0[9] + 6.0 * x2 * y2 * z5 * r0[10] +
    60.0 * x2 * y2 * z3 * r0[9] + 90.0 * x2 * y2 * z * r0[8] +
    3.0 * x2 * z5 * r0[9] + 30.0 * x2 * z3 * r0[8] + 45.0 * x2 * z * r0[7] +
    y4 * z5 * r0[10] + 10.0 * y4 * z3 * r0[9] + 15.0 * y4 * z * r0[8] +
    6.0 * y2 * z5 * r0[9] + 60.0 * y2 * z3 * r0[8] + 90.0 * y2 * z * r0[7] +
    3.0 * z5 * r0[8] + 30.0 * z3 * r0[7] + 45.0 * z * r0[6];
  r[3][3][5] =
    x3 * y3 * z5 * r0[11] + 10.0 * x3 * y3 * z3 * r0[10] +
    15.0 * x3 * y3 * z * r0[9] + 3.0 * x3 * y * z5 * r0[10] +
    30.0 * x3 * y * z3 * r0[9] + 45.0 * x3 * y * z * r0[8] +
    3.0 * x * y3 * z5 * r0[10] + 30.0 * x * y3 * z3 * r0[9] +
    45.0 * x * y3 * z * r0[8] + 9.0 * x * y * z5 * r0[9] +
    90.0 * x * y * z3 * r0[8] + 135.0 * x * y * z * r0[7];
  r[4][2][5] =
    x4 * y2 * z5 * r0[11] + 10.0 * x4 * y2 * z3 * r0[10] +
    15.0 * x4 * y2 * z * r0[9] + x4 * z5 * r0[10] + 10.0 * x4 * z3 * r0[9] +
    15.0 * x4 * z * r0[8] + 6.0 * x2 * y2 * z5 * r0[10] +
    60.0 * x2 * y2 * z3 * r0[9] + 90.0 * x2 * y2 * z * r0[8] +
    6.0 * x2 * z5 * r0[9] + 60.0 * x2 * z3 * r0[8] + 90.0 * x2 * z * r0[7] +
    3.0 * y2 * z5 * r0[9] + 30.0 * y2 * z3 * r0[8] + 45.0 * y2 * z * r0[7] +
    3.0 * z5 * r0[8] + 30.0 * z3 * r0[7] + 45.0 * z * r0[6];
  r[5][1][5] =
    x5 * y * z5 * r0[11] + 10.0 * x5 * y * z3 * r0[10] +
    15.0 * x5 * y * z * r0[9] + 10.0 * x3 * y * z5 * r0[10] +
    100.0 * x3 * y * z3 * r0[9] + 150.0 * x3 * y * z * r0[8] +
    15.0 * x * y * z5 * r0[9] + 150.0 * x * y * z3 * r0[8] +
    225.0 * x * y * z * r0[7];
  r[6][0][5] =
    x6 * z5 * r0[11] + 10.0 * x6 * z3 * r0[10] + 15.0 * x6 * z * r0[9] +
    15.0 * x4 * z5 * r0[10] + 150.0 * x4 * z3 * r0[9] +
    225.0 * x4 * z * r0[8] + 45.0 * x2 * z5 * r0[9] +
    450.0 * x2 * z3 * r0[8] + 675.0 * x2 * z * r0[7] + 15.0 * z5 * r0[8] +
    150.0 * z3 * r0[7] + 225.0 * z * r0[6];
  r[0][7][4] =
    y7 * z4 * r0[11] + 6.0 * y7 * z2 * r0[10] + 3.0 * y7 * r0[9] +
    21.0 * y5 * z4 * r0[10] + 126.0 * y5 * z2 * r0[9] + 63.0 * y5 * r0[8] +
    105.0 * y3 * z4 * r0[9] + 630.0 * y3 * z2 * r0[8] + 315.0 * y3 * r0[7] +
    105.0 * y * z4 * r0[8] + 630.0 * y * z2 * r0[7] + 315.0 * y * r0[6];
  r[1][6][4] =
    x * y6 * z4 * r0[11] + 6.0 * x * y6 * z2 * r0[10] + 3.0 * x * y6 * r0[9] +
    15.0 * x * y4 * z4 * r0[10] + 90.0 * x * y4 * z2 * r0[9] +
    45.0 * x * y4 * r0[8] + 45.0 * x * y2 * z4 * r0[9] +
    270.0 * x * y2 * z2 * r0[8] + 135.0 * x * y2 * r0[7] +
    15.0 * x * z4 * r0[8] + 90.0 * x * z2 * r0[7] + 45.0 * x * r0[6];
  r[2][5][4] =
    x2 * y5 * z4 * r0[11] + 6.0 * x2 * y5 * z2 * r0[10] +
    3.0 * x2 * y5 * r0[9] + 10.0 * x2 * y3 * z4 * r0[10] +
    60.0 * x2 * y3 * z2 * r0[9] + 30.0 * x2 * y3 * r0[8] +
    15.0 * x2 * y * z4 * r0[9] + 90.0 * x2 * y * z2 * r0[8] +
    45.0 * x2 * y * r0[7] + y5 * z4 * r0[10] + 6.0 * y5 * z2 * r0[9] +
    3.0 * y5 * r0[8] + 10.0 * y3 * z4 * r0[9] + 60.0 * y3 * z2 * r0[8] +
    30.0 * y3 * r0[7] + 15.0 * y * z4 * r0[8] + 90.0 * y * z2 * r0[7] +
    45.0 * y * r0[6];
  r[3][4][4] =
    x3 * y4 * z4 * r0[11] + 6.0 * x3 * y4 * z2 * r0[10] +
    3.0 * x3 * y4 * r0[9] + 6.0 * x3 * y2 * z4 * r0[10] +
    36.0 * x3 * y2 * z2 * r0[9] + 18.0 * x3 * y2 * r0[8] +
    3.0 * x3 * z4 * r0[9] + 18.0 * x3 * z2 * r0[8] + 9.0 * x3 * r0[7] +
    3.0 * x * y4 * z4 * r0[10] + 18.0 * x * y4 * z2 * r0[9] +
    9.0 * x * y4 * r0[8] + 18.0 * x * y2 * z4 * r0[9] +
    108.0 * x * y2 * z2 * r0[8] + 54.0 * x * y2 * r0[7] +
    9.0 * x * z4 * r0[8] + 54.0 * x * z2 * r0[7] + 27.0 * x * r0[6];
  r[4][3][4] =
    x4 * y3 * z4 * r0[11] + 6.0 * x4 * y3 * z2 * r0[10] +
    3.0 * x4 * y3 * r0[9] + 3.0 * x4 * y * z4 * r0[10] +
    18.0 * x4 * y * z2 * r0[9] + 9.0 * x4 * y * r0[8] +
    6.0 * x2 * y3 * z4 * r0[10] + 36.0 * x2 * y3 * z2 * r0[9] +
    18.0 * x2 * y3 * r0[8] + 18.0 * x2 * y * z4 * r0[9] +
    108.0 * x2 * y * z2 * r0[8] + 54.0 * x2 * y * r0[7] +
    3.0 * y3 * z4 * r0[9] + 18.0 * y3 * z2 * r0[8] + 9.0 * y3 * r0[7] +
    9.0 * y * z4 * r0[8] + 54.0 * y * z2 * r0[7] + 27.0 * y * r0[6];
  r[5][2][4] =
    x5 * y2 * z4 * r0[11] + 6.0 * x5 * y2 * z2 * r0[10] +
    3.0 * x5 * y2 * r0[9] + x5 * z4 * r0[10] + 6.0 * x5 * z2 * r0[9] +
    3.0 * x5 * r0[8] + 10.0 * x3 * y2 * z4 * r0[10] +
    60.0 * x3 * y2 * z2 * r0[9] + 30.0 * x3 * y2 * r0[8] +
    10.0 * x3 * z4 * r0[9] + 60.0 * x3 * z2 * r0[8] + 30.0 * x3 * r0[7] +
    15.0 * x * y2 * z4 * r0[9] + 90.0 * x * y2 * z2 * r0[8] +
    45.0 * x * y2 * r0[7] + 15.0 * x * z4 * r0[8] + 90.0 * x * z2 * r0[7] +
    45.0 * x * r0[6];
  r[6][1][4] =
    x6 * y * z4 * r0[11] + 6.0 * x6 * y * z2 * r0[10] + 3.0 * x6 * y * r0[9] +
    15.0 * x4 * y * z4 * r0[10] + 90.0 * x4 * y * z2 * r0[9] +
    45.0 * x4 * y * r0[8] + 45.0 * x2 * y * z4 * r0[9] +
    270.0 * x2 * y * z2 * r0[8] + 135.0 * x2 * y * r0[7] +
    15.0 * y * z4 * r0[8] + 90.0 * y * z2 * r0[7] + 45.0 * y * r0[6];
  r[7][0][4] =
    x7 * z4 * r0[11] + 6.0 * x7 * z2 * r0[10] + 3.0 * x7 * r0[9] +
    21.0 * x5 * z4 * r0[10] + 126.0 * x5 * z2 * r0[9] + 63.0 * x5 * r0[8] +
    105.0 * x3 * z4 * r0[9] + 630.0 * x3 * z2 * r0[8] + 315.0 * x3 * r0[7] +
    105.0 * x * z4 * r0[8] + 630.0 * x * z2 * r0[7] + 315.0 * x * r0[6];
  r[0][8][3] =
    y8 * z3 * r0[11] + 3.0 * y8 * z * r0[10] + 28.0 * y6 * z3 * r0[10] +
    84.0 * y6 * z * r0[9] + 210.0 * y4 * z3 * r0[9] + 630.0 * y4 * z * r0[8] +
    420.0 * y2 * z3 * r0[8] + 1260.0 * y2 * z * r0[7] + 105.0 * z3 * r0[7] +
    315.0 * z * r0[6];
  r[1][7][3] =
    x * y7 * z3 * r0[11] + 3.0 * x * y7 * z * r0[10] +
    21.0 * x * y5 * z3 * r0[10] + 63.0 * x * y5 * z * r0[9] +
    105.0 * x * y3 * z3 * r0[9] + 315.0 * x * y3 * z * r0[8] +
    105.0 * x * y * z3 * r0[8] + 315.0 * x * y * z * r0[7];
  r[2][6][3] =
    x2 * y6 * z3 * r0[11] + 3.0 * x2 * y6 * z * r0[10] +
    15.0 * x2 * y4 * z3 * r0[10] + 45.0 * x2 * y4 * z * r0[9] +
    45.0 * x2 * y2 * z3 * r0[9] + 135.0 * x2 * y2 * z * r0[8] +
    15.0 * x2 * z3 * r0[8] + 45.0 * x2 * z * r0[7] + y6 * z3 * r0[10] +
    3.0 * y6 * z * r0[9] + 15.0 * y4 * z3 * r0[9] + 45.0 * y4 * z * r0[8] +
    45.0 * y2 * z3 * r0[8] + 135.0 * y2 * z * r0[7] + 15.0 * z3 * r0[7] +
    45.0 * z * r0[6];
  r[3][5][3] =
    x3 * y5 * z3 * r0[11] + 3.0 * x3 * y5 * z * r0[10] +
    10.0 * x3 * y3 * z3 * r0[10] + 30.0 * x3 * y3 * z * r0[9] +
    15.0 * x3 * y * z3 * r0[9] + 45.0 * x3 * y * z * r0[8] +
    3.0 * x * y5 * z3 * r0[10] + 9.0 * x * y5 * z * r0[9] +
    30.0 * x * y3 * z3 * r0[9] + 90.0 * x * y3 * z * r0[8] +
    45.0 * x * y * z3 * r0[8] + 135.0 * x * y * z * r0[7];
  r[4][4][3] =
    x4 * y4 * z3 * r0[11] + 3.0 * x4 * y4 * z * r0[10] +
    6.0 * x4 * y2 * z3 * r0[10] + 18.0 * x4 * y2 * z * r0[9] +
    3.0 * x4 * z3 * r0[9] + 9.0 * x4 * z * r0[8] +
    6.0 * x2 * y4 * z3 * r0[10] + 18.0 * x2 * y4 * z * r0[9] +
    36.0 * x2 * y2 * z3 * r0[9] + 108.0 * x2 * y2 * z * r0[8] +
    18.0 * x2 * z3 * r0[8] + 54.0 * x2 * z * r0[7] + 3.0 * y4 * z3 * r0[9] +
    9.0 * y4 * z * r0[8] + 18.0 * y2 * z3 * r0[8] + 54.0 * y2 * z * r0[7] +
    9.0 * z3 * r0[7] + 27.0 * z * r0[6];
  r[5][3][3] =
    x5 * y3 * z3 * r0[11] + 3.0 * x5 * y3 * z * r0[10] +
    3.0 * x5 * y * z3 * r0[10] + 9.0 * x5 * y * z * r0[9] +
    10.0 * x3 * y3 * z3 * r0[10] + 30.0 * x3 * y3 * z * r0[9] +
    30.0 * x3 * y * z3 * r0[9] + 90.0 * x3 * y * z * r0[8] +
    15.0 * x * y3 * z3 * r0[9] + 45.0 * x * y3 * z * r0[8] +
    45.0 * x * y * z3 * r0[8] + 135.0 * x * y * z * r0[7];
  r[6][2][3] =
    x6 * y2 * z3 * r0[11] + 3.0 * x6 * y2 * z * r0[10] + x6 * z3 * r0[10] +
    3.0 * x6 * z * r0[9] + 15.0 * x4 * y2 * z3 * r0[10] +
    45.0 * x4 * y2 * z * r0[9] + 15.0 * x4 * z3 * r0[9] +
    45.0 * x4 * z * r0[8] + 45.0 * x2 * y2 * z3 * r0[9] +
    135.0 * x2 * y2 * z * r0[8] + 45.0 * x2 * z3 * r0[8] +
    135.0 * x2 * z * r0[7] + 15.0 * y2 * z3 * r0[8] + 45.0 * y2 * z * r0[7] +
    15.0 * z3 * r0[7] + 45.0 * z * r0[6];
  r[7][1][3] =
    x7 * y * z3 * r0[11] + 3.0 * x7 * y * z * r0[10] +
    21.0 * x5 * y * z3 * r0[10] + 63.0 * x5 * y * z * r0[9] +
    105.0 * x3 * y * z3 * r0[9] + 315.0 * x3 * y * z * r0[8] +
    105.0 * x * y * z3 * r0[8] + 315.0 * x * y * z * r0[7];
  r[8][0][3] =
    x8 * z3 * r0[11] + 3.0 * x8 * z * r0[10] + 28.0 * x6 * z3 * r0[10] +
    84.0 * x6 * z * r0[9] + 210.0 * x4 * z3 * r0[9] + 630.0 * x4 * z * r0[8] +
    420.0 * x2 * z3 * r0[8] + 1260.0 * x2 * z * r0[7] + 105.0 * z3 * r0[7] +
    315.0 * z * r0[6];
  r[0][9][2] =
    y9 * z2 * r0[11] + y9 * r0[10] + 36.0 * y7 * z2 * r0[10] +
    36.0 * y7 * r0[9] + 378.0 * y5 * z2 * r0[9] + 378.0 * y5 * r0[8] +
    1260.0 * y3 * z2 * r0[8] + 1260.0 * y3 * r0[7] + 945.0 * y * z2 * r0[7] +
    945.0 * y * r0[6];
  r[1][8][2] =
    x * y8 * z2 * r0[11] + x * y8 * r0[10] + 28.0 * x * y6 * z2 * r0[10] +
    28.0 * x * y6 * r0[9] + 210.0 * x * y4 * z2 * r0[9] +
    210.0 * x * y4 * r0[8] + 420.0 * x * y2 * z2 * r0[8] +
    420.0 * x * y2 * r0[7] + 105.0 * x * z2 * r0[7] + 105.0 * x * r0[6];
  r[2][7][2] =
    x2 * y7 * z2 * r0[11] + x2 * y7 * r0[10] + 21.0 * x2 * y5 * z2 * r0[10] +
    21.0 * x2 * y5 * r0[9] + 105.0 * x2 * y3 * z2 * r0[9] +
    105.0 * x2 * y3 * r0[8] + 105.0 * x2 * y * z2 * r0[8] +
    105.0 * x2 * y * r0[7] + y7 * z2 * r0[10] + y7 * r0[9] +
    21.0 * y5 * z2 * r0[9] + 21.0 * y5 * r0[8] + 105.0 * y3 * z2 * r0[8] +
    105.0 * y3 * r0[7] + 105.0 * y * z2 * r0[7] + 105.0 * y * r0[6];
  r[3][6][2] =
    x3 * y6 * z2 * r0[11] + x3 * y6 * r0[10] + 15.0 * x3 * y4 * z2 * r0[10] +
    15.0 * x3 * y4 * r0[9] + 45.0 * x3 * y2 * z2 * r0[9] +
    45.0 * x3 * y2 * r0[8] + 15.0 * x3 * z2 * r0[8] + 15.0 * x3 * r0[7] +
    3.0 * x * y6 * z2 * r0[10] + 3.0 * x * y6 * r0[9] +
    45.0 * x * y4 * z2 * r0[9] + 45.0 * x * y4 * r0[8] +
    135.0 * x * y2 * z2 * r0[8] + 135.0 * x * y2 * r0[7] +
    45.0 * x * z2 * r0[7] + 45.0 * x * r0[6];
  r[4][5][2] =
    x4 * y5 * z2 * r0[11] + x4 * y5 * r0[10] + 10.0 * x4 * y3 * z2 * r0[10] +
    10.0 * x4 * y3 * r0[9] + 15.0 * x4 * y * z2 * r0[9] +
    15.0 * x4 * y * r0[8] + 6.0 * x2 * y5 * z2 * r0[10] +
    6.0 * x2 * y5 * r0[9] + 60.0 * x2 * y3 * z2 * r0[9] +
    60.0 * x2 * y3 * r0[8] + 90.0 * x2 * y * z2 * r0[8] +
    90.0 * x2 * y * r0[7] + 3.0 * y5 * z2 * r0[9] + 3.0 * y5 * r0[8] +
    30.0 * y3 * z2 * r0[8] + 30.0 * y3 * r0[7] + 45.0 * y * z2 * r0[7] +
    45.0 * y * r0[6];
  r[5][4][2] =
    x5 * y4 * z2 * r0[11] + x5 * y4 * r0[10] + 6.0 * x5 * y2 * z2 * r0[10] +
    6.0 * x5 * y2 * r0[9] + 3.0 * x5 * z2 * r0[9] + 3.0 * x5 * r0[8] +
    10.0 * x3 * y4 * z2 * r0[10] + 10.0 * x3 * y4 * r0[9] +
    60.0 * x3 * y2 * z2 * r0[9] + 60.0 * x3 * y2 * r0[8] +
    30.0 * x3 * z2 * r0[8] + 30.0 * x3 * r0[7] + 15.0 * x * y4 * z2 * r0[9] +
    15.0 * x * y4 * r0[8] + 90.0 * x * y2 * z2 * r0[8] +
    90.0 * x * y2 * r0[7] + 45.0 * x * z2 * r0[7] + 45.0 * x * r0[6];
  r[6][3][2] =
    x6 * y3 * z2 * r0[11] + x6 * y3 * r0[10] + 3.0 * x6 * y * z2 * r0[10] +
    3.0 * x6 * y * r0[9] + 15.0 * x4 * y3 * z2 * r0[10] +
    15.0 * x4 * y3 * r0[9] + 45.0 * x4 * y * z2 * r0[9] +
    45.0 * x4 * y * r0[8] + 45.0 * x2 * y3 * z2 * r0[9] +
    45.0 * x2 * y3 * r0[8] + 135.0 * x2 * y * z2 * r0[8] +
    135.0 * x2 * y * r0[7] + 15.0 * y3 * z2 * r0[8] + 15.0 * y3 * r0[7] +
    45.0 * y * z2 * r0[7] + 45.0 * y * r0[6];
  r[7][2][2] =
    x7 * y2 * z2 * r0[11] + x7 * y2 * r0[10] + x7 * z2 * r0[10] + x7 * r0[9] +
    21.0 * x5 * y2 * z2 * r0[10] + 21.0 * x5 * y2 * r0[9] +
    21.0 * x5 * z2 * r0[9] + 21.0 * x5 * r0[8] +
    105.0 * x3 * y2 * z2 * r0[9] + 105.0 * x3 * y2 * r0[8] +
    105.0 * x3 * z2 * r0[8] + 105.0 * x3 * r0[7] +
    105.0 * x * y2 * z2 * r0[8] + 105.0 * x * y2 * r0[7] +
    105.0 * x * z2 * r0[7] + 105.0 * x * r0[6];
  r[8][1][2] =
    x8 * y * z2 * r0[11] + x8 * y * r0[10] + 28.0 * x6 * y * z2 * r0[10] +
    28.0 * x6 * y * r0[9] + 210.0 * x4 * y * z2 * r0[9] +
    210.0 * x4 * y * r0[8] + 420.0 * x2 * y * z2 * r0[8] +
    420.0 * x2 * y * r0[7] + 105.0 * y * z2 * r0[7] + 105.0 * y * r0[6];
  r[9][0][2] =
    x9 * z2 * r0[11] + x9 * r0[10] + 36.0 * x7 * z2 * r0[10] +
    36.0 * x7 * r0[9] + 378.0 * x5 * z2 * r0[9] + 378.0 * x5 * r0[8] +
    1260.0 * x3 * z2 * r0[8] + 1260.0 * x3 * r0[7] + 945.0 * x * z2 * r0[7] +
    945.0 * x * r0[6];
  r[0][10][1] =
    y10 * z * r0[11] + 45.0 * y8 * z * r0[10] + 630.0 * y6 * z * r0[9] +
    3150.0 * y4 * z * r0[8] + 4725.0 * y2 * z * r0[7] + 945.0 * z * r0[6];
  r[1][9][1] =
    x * y9 * z * r0[11] + 36.0 * x * y7 * z * r0[10] +
    378.0 * x * y5 * z * r0[9] + 1260.0 * x * y3 * z * r0[8] +
    945.0 * x * y * z * r0[7];
  r[2][8][1] =
    x2 * y8 * z * r0[11] + 28.0 * x2 * y6 * z * r0[10] +
    210.0 * x2 * y4 * z * r0[9] + 420.0 * x2 * y2 * z * r0[8] +
    105.0 * x2 * z * r0[7] + y8 * z * r0[10] + 28.0 * y6 * z * r0[9] +
    210.0 * y4 * z * r0[8] + 420.0 * y2 * z * r0[7] + 105.0 * z * r0[6];
  r[3][7][1] =
    x3 * y7 * z * r0[11] + 21.0 * x3 * y5 * z * r0[10] +
    105.0 * x3 * y3 * z * r0[9] + 105.0 * x3 * y * z * r0[8] +
    3.0 * x * y7 * z * r0[10] + 63.0 * x * y5 * z * r0[9] +
    315.0 * x * y3 * z * r0[8] + 315.0 * x * y * z * r0[7];
  r[4][6][1] =
    x4 * y6 * z * r0[11] + 15.0 * x4 * y4 * z * r0[10] +
    45.0 * x4 * y2 * z * r0[9] + 15.0 * x4 * z * r0[8] +
    6.0 * x2 * y6 * z * r0[10] + 90.0 * x2 * y4 * z * r0[9] +
    270.0 * x2 * y2 * z * r0[8] + 90.0 * x2 * z * r0[7] +
    3.0 * y6 * z * r0[9] + 45.0 * y4 * z * r0[8] + 135.0 * y2 * z * r0[7] +
    45.0 * z * r0[6];
  r[5][5][1] =
    x5 * y5 * z * r0[11] + 10.0 * x5 * y3 * z * r0[10] +
    15.0 * x5 * y * z * r0[9] + 10.0 * x3 * y5 * z * r0[10] +
    100.0 * x3 * y3 * z * r0[9] + 150.0 * x3 * y * z * r0[8] +
    15.0 * x * y5 * z * r0[9] + 150.0 * x * y3 * z * r0[8] +
    225.0 * x * y * z * r0[7];
  r[6][4][1] =
    x6 * y4 * z * r0[11] + 6.0 * x6 * y2 * z * r0[10] + 3.0 * x6 * z * r0[9] +
    15.0 * x4 * y4 * z * r0[10] + 90.0 * x4 * y2 * z * r0[9] +
    45.0 * x4 * z * r0[8] + 45.0 * x2 * y4 * z * r0[9] +
    270.0 * x2 * y2 * z * r0[8] + 135.0 * x2 * z * r0[7] +
    15.0 * y4 * z * r0[8] + 90.0 * y2 * z * r0[7] + 45.0 * z * r0[6];
  r[7][3][1] =
    x7 * y3 * z * r0[11] + 3.0 * x7 * y * z * r0[10] +
    21.0 * x5 * y3 * z * r0[10] + 63.0 * x5 * y * z * r0[9] +
    105.0 * x3 * y3 * z * r0[9] + 315.0 * x3 * y * z * r0[8] +
    105.0 * x * y3 * z * r0[8] + 315.0 * x * y * z * r0[7];
  r[8][2][1] =
    x8 * y2 * z * r0[11] + x8 * z * r0[10] + 28.0 * x6 * y2 * z * r0[10] +
    28.0 * x6 * z * r0[9] + 210.0 * x4 * y2 * z * r0[9] +
    210.0 * x4 * z * r0[8] + 420.0 * x2 * y2 * z * r0[8] +
    420.0 * x2 * z * r0[7] + 105.0 * y2 * z * r0[7] + 105.0 * z * r0[6];
  r[9][1][1] =
    x9 * y * z * r0[11] + 36.0 * x7 * y * z * r0[10] +
    378.0 * x5 * y * z * r0[9] + 1260.0 * x3 * y * z * r0[8] +
    945.0 * x * y * z * r0[7];
  r[10][0][1] =
    x10 * z * r0[11] + 45.0 * x8 * z * r0[10] + 630.0 * x6 * z * r0[9] +
    3150.0 * x4 * z * r0[8] + 4725.0 * x2 * z * r0[7] + 945.0 * z * r0[6];
  r[0][11][0] =
    y11 * r0[11] + 55.0 * y9 * r0[10] + 990.0 * y7 * r0[9] +
    6930.0 * y5 * r0[8] + 17325.0 * y3 * r0[7] + 10395.0 * y * r0[6];
  r[1][10][0] =
    x * y10 * r0[11] + 45.0 * x * y8 * r0[10] + 630.0 * x * y6 * r0[9] +
    3150.0 * x * y4 * r0[8] + 4725.0 * x * y2 * r0[7] + 945.0 * x * r0[6];
  r[2][9][0] =
    x2 * y9 * r0[11] + 36.0 * x2 * y7 * r0[10] + 378.0 * x2 * y5 * r0[9] +
    1260.0 * x2 * y3 * r0[8] + 945.0 * x2 * y * r0[7] + y9 * r0[10] +
    36.0 * y7 * r0[9] + 378.0 * y5 * r0[8] + 1260.0 * y3 * r0[7] +
    945.0 * y * r0[6];
  r[3][8][0] =
    x3 * y8 * r0[11] + 28.0 * x3 * y6 * r0[10] + 210.0 * x3 * y4 * r0[9] +
    420.0 * x3 * y2 * r0[8] + 105.0 * x3 * r0[7] + 3.0 * x * y8 * r0[10] +
    84.0 * x * y6 * r0[9] + 630.0 * x * y4 * r0[8] + 1260.0 * x * y2 * r0[7] +
    315.0 * x * r0[6];
  r[4][7][0] =
    x4 * y7 * r0[11] + 21.0 * x4 * y5 * r0[10] + 105.0 * x4 * y3 * r0[9] +
    105.0 * x4 * y * r0[8] + 6.0 * x2 * y7 * r0[10] +
    126.0 * x2 * y5 * r0[9] + 630.0 * x2 * y3 * r0[8] +
    630.0 * x2 * y * r0[7] + 3.0 * y7 * r0[9] + 63.0 * y5 * r0[8] +
    315.0 * y3 * r0[7] + 315.0 * y * r0[6];
  r[5][6][0] =
    x5 * y6 * r0[11] + 15.0 * x5 * y4 * r0[10] + 45.0 * x5 * y2 * r0[9] +
    15.0 * x5 * r0[8] + 10.0 * x3 * y6 * r0[10] + 150.0 * x3 * y4 * r0[9] +
    450.0 * x3 * y2 * r0[8] + 150.0 * x3 * r0[7] + 15.0 * x * y6 * r0[9] +
    225.0 * x * y4 * r0[8] + 675.0 * x * y2 * r0[7] + 225.0 * x * r0[6];
  r[6][5][0] =
    x6 * y5 * r0[11] + 10.0 * x6 * y3 * r0[10] + 15.0 * x6 * y * r0[9] +
    15.0 * x4 * y5 * r0[10] + 150.0 * x4 * y3 * r0[9] +
    225.0 * x4 * y * r0[8] + 45.0 * x2 * y5 * r0[9] +
    450.0 * x2 * y3 * r0[8] + 675.0 * x2 * y * r0[7] + 15.0 * y5 * r0[8] +
    150.0 * y3 * r0[7] + 225.0 * y * r0[6];
  r[7][4][0] =
    x7 * y4 * r0[11] + 6.0 * x7 * y2 * r0[10] + 3.0 * x7 * r0[9] +
    21.0 * x5 * y4 * r0[10] + 126.0 * x5 * y2 * r0[9] + 63.0 * x5 * r0[8] +
    105.0 * x3 * y4 * r0[9] + 630.0 * x3 * y2 * r0[8] + 315.0 * x3 * r0[7] +
    105.0 * x * y4 * r0[8] + 630.0 * x * y2 * r0[7] + 315.0 * x * r0[6];
  r[8][3][0] =
    x8 * y3 * r0[11] + 3.0 * x8 * y * r0[10] + 28.0 * x6 * y3 * r0[10] +
    84.0 * x6 * y * r0[9] + 210.0 * x4 * y3 * r0[9] + 630.0 * x4 * y * r0[8] +
    420.0 * x2 * y3 * r0[8] + 1260.0 * x2 * y * r0[7] + 105.0 * y3 * r0[7] +
    315.0 * y * r0[6];
  r[9][2][0] =
    x9 * y2 * r0[11] + x9 * r0[10] + 36.0 * x7 * y2 * r0[10] +
    36.0 * x7 * r0[9] + 378.0 * x5 * y2 * r0[9] + 378.0 * x5 * r0[8] +
    1260.0 * x3 * y2 * r0[8] + 1260.0 * x3 * r0[7] + 945.0 * x * y2 * r0[7] +
    945.0 * x * r0[6];
  r[10][1][0] =
    x10 * y * r0[11] + 45.0 * x8 * y * r0[10] + 630.0 * x6 * y * r0[9] +
    3150.0 * x4 * y * r0[8] + 4725.0 * x2 * y * r0[7] + 945.0 * y * r0[6];
  r[11][0][0] =
    x11 * r0[11] + 55.0 * x9 * r0[10] + 990.0 * x7 * r0[9] +
    6930.0 * x5 * r0[8] + 17325.0 * x3 * r0[7] + 10395.0 * x * r0[6];
  if (ltot == 11)
    {
      return;
    }
  x12 = x11 * x;
  y12 = y11 * y;
  z12 = z11 * z;
  r[0][0][12] =
    z12 * r0[12] + 66.0 * z10 * r0[11] + 1485.0 * z8 * r0[10] +
    13860.0 * z6 * r0[9] + 51975.0 * z4 * r0[8] + 62370.0 * z2 * r0[7] +
    10395.0 * r0[6];
  r[0][1][11] =
    y * z11 * r0[12] + 55.0 * y * z9 * r0[11] + 990.0 * y * z7 * r0[10] +
    6930.0 * y * z5 * r0[9] + 17325.0 * y * z3 * r0[8] +
    10395.0 * y * z * r0[7];
  r[1][0][11] =
    x * z11 * r0[12] + 55.0 * x * z9 * r0[11] + 990.0 * x * z7 * r0[10] +
    6930.0 * x * z5 * r0[9] + 17325.0 * x * z3 * r0[8] +
    10395.0 * x * z * r0[7];
  r[0][2][10] =
    y2 * z10 * r0[12] + 45.0 * y2 * z8 * r0[11] + 630.0 * y2 * z6 * r0[10] +
    3150.0 * y2 * z4 * r0[9] + 4725.0 * y2 * z2 * r0[8] + 945.0 * y2 * r0[7] +
    z10 * r0[11] + 45.0 * z8 * r0[10] + 630.0 * z6 * r0[9] +
    3150.0 * z4 * r0[8] + 4725.0 * z2 * r0[7] + 945.0 * r0[6];
  r[1][1][10] =
    x * y * z10 * r0[12] + 45.0 * x * y * z8 * r0[11] +
    630.0 * x * y * z6 * r0[10] + 3150.0 * x * y * z4 * r0[9] +
    4725.0 * x * y * z2 * r0[8] + 945.0 * x * y * r0[7];
  r[2][0][10] =
    x2 * z10 * r0[12] + 45.0 * x2 * z8 * r0[11] + 630.0 * x2 * z6 * r0[10] +
    3150.0 * x2 * z4 * r0[9] + 4725.0 * x2 * z2 * r0[8] + 945.0 * x2 * r0[7] +
    z10 * r0[11] + 45.0 * z8 * r0[10] + 630.0 * z6 * r0[9] +
    3150.0 * z4 * r0[8] + 4725.0 * z2 * r0[7] + 945.0 * r0[6];
  r[0][3][9] =
    y3 * z9 * r0[12] + 36.0 * y3 * z7 * r0[11] + 378.0 * y3 * z5 * r0[10] +
    1260.0 * y3 * z3 * r0[9] + 945.0 * y3 * z * r0[8] +
    3.0 * y * z9 * r0[11] + 108.0 * y * z7 * r0[10] +
    1134.0 * y * z5 * r0[9] + 3780.0 * y * z3 * r0[8] +
    2835.0 * y * z * r0[7];
  r[1][2][9] =
    x * y2 * z9 * r0[12] + 36.0 * x * y2 * z7 * r0[11] +
    378.0 * x * y2 * z5 * r0[10] + 1260.0 * x * y2 * z3 * r0[9] +
    945.0 * x * y2 * z * r0[8] + x * z9 * r0[11] + 36.0 * x * z7 * r0[10] +
    378.0 * x * z5 * r0[9] + 1260.0 * x * z3 * r0[8] + 945.0 * x * z * r0[7];
  r[2][1][9] =
    x2 * y * z9 * r0[12] + 36.0 * x2 * y * z7 * r0[11] +
    378.0 * x2 * y * z5 * r0[10] + 1260.0 * x2 * y * z3 * r0[9] +
    945.0 * x2 * y * z * r0[8] + y * z9 * r0[11] + 36.0 * y * z7 * r0[10] +
    378.0 * y * z5 * r0[9] + 1260.0 * y * z3 * r0[8] + 945.0 * y * z * r0[7];
  r[3][0][9] =
    x3 * z9 * r0[12] + 36.0 * x3 * z7 * r0[11] + 378.0 * x3 * z5 * r0[10] +
    1260.0 * x3 * z3 * r0[9] + 945.0 * x3 * z * r0[8] +
    3.0 * x * z9 * r0[11] + 108.0 * x * z7 * r0[10] +
    1134.0 * x * z5 * r0[9] + 3780.0 * x * z3 * r0[8] +
    2835.0 * x * z * r0[7];
  r[0][4][8] =
    y4 * z8 * r0[12] + 28.0 * y4 * z6 * r0[11] + 210.0 * y4 * z4 * r0[10] +
    420.0 * y4 * z2 * r0[9] + 105.0 * y4 * r0[8] + 6.0 * y2 * z8 * r0[11] +
    168.0 * y2 * z6 * r0[10] + 1260.0 * y2 * z4 * r0[9] +
    2520.0 * y2 * z2 * r0[8] + 630.0 * y2 * r0[7] + 3.0 * z8 * r0[10] +
    84.0 * z6 * r0[9] + 630.0 * z4 * r0[8] + 1260.0 * z2 * r0[7] +
    315.0 * r0[6];
  r[1][3][8] =
    x * y3 * z8 * r0[12] + 28.0 * x * y3 * z6 * r0[11] +
    210.0 * x * y3 * z4 * r0[10] + 420.0 * x * y3 * z2 * r0[9] +
    105.0 * x * y3 * r0[8] + 3.0 * x * y * z8 * r0[11] +
    84.0 * x * y * z6 * r0[10] + 630.0 * x * y * z4 * r0[9] +
    1260.0 * x * y * z2 * r0[8] + 315.0 * x * y * r0[7];
  r[2][2][8] =
    x2 * y2 * z8 * r0[12] + 28.0 * x2 * y2 * z6 * r0[11] +
    210.0 * x2 * y2 * z4 * r0[10] + 420.0 * x2 * y2 * z2 * r0[9] +
    105.0 * x2 * y2 * r0[8] + x2 * z8 * r0[11] + 28.0 * x2 * z6 * r0[10] +
    210.0 * x2 * z4 * r0[9] + 420.0 * x2 * z2 * r0[8] + 105.0 * x2 * r0[7] +
    y2 * z8 * r0[11] + 28.0 * y2 * z6 * r0[10] + 210.0 * y2 * z4 * r0[9] +
    420.0 * y2 * z2 * r0[8] + 105.0 * y2 * r0[7] + z8 * r0[10] +
    28.0 * z6 * r0[9] + 210.0 * z4 * r0[8] + 420.0 * z2 * r0[7] +
    105.0 * r0[6];
  r[3][1][8] =
    x3 * y * z8 * r0[12] + 28.0 * x3 * y * z6 * r0[11] +
    210.0 * x3 * y * z4 * r0[10] + 420.0 * x3 * y * z2 * r0[9] +
    105.0 * x3 * y * r0[8] + 3.0 * x * y * z8 * r0[11] +
    84.0 * x * y * z6 * r0[10] + 630.0 * x * y * z4 * r0[9] +
    1260.0 * x * y * z2 * r0[8] + 315.0 * x * y * r0[7];
  r[4][0][8] =
    x4 * z8 * r0[12] + 28.0 * x4 * z6 * r0[11] + 210.0 * x4 * z4 * r0[10] +
    420.0 * x4 * z2 * r0[9] + 105.0 * x4 * r0[8] + 6.0 * x2 * z8 * r0[11] +
    168.0 * x2 * z6 * r0[10] + 1260.0 * x2 * z4 * r0[9] +
    2520.0 * x2 * z2 * r0[8] + 630.0 * x2 * r0[7] + 3.0 * z8 * r0[10] +
    84.0 * z6 * r0[9] + 630.0 * z4 * r0[8] + 1260.0 * z2 * r0[7] +
    315.0 * r0[6];
  r[0][5][7] =
    y5 * z7 * r0[12] + 21.0 * y5 * z5 * r0[11] + 105.0 * y5 * z3 * r0[10] +
    105.0 * y5 * z * r0[9] + 10.0 * y3 * z7 * r0[11] +
    210.0 * y3 * z5 * r0[10] + 1050.0 * y3 * z3 * r0[9] +
    1050.0 * y3 * z * r0[8] + 15.0 * y * z7 * r0[10] +
    315.0 * y * z5 * r0[9] + 1575.0 * y * z3 * r0[8] + 1575.0 * y * z * r0[7];
  r[1][4][7] =
    x * y4 * z7 * r0[12] + 21.0 * x * y4 * z5 * r0[11] +
    105.0 * x * y4 * z3 * r0[10] + 105.0 * x * y4 * z * r0[9] +
    6.0 * x * y2 * z7 * r0[11] + 126.0 * x * y2 * z5 * r0[10] +
    630.0 * x * y2 * z3 * r0[9] + 630.0 * x * y2 * z * r0[8] +
    3.0 * x * z7 * r0[10] + 63.0 * x * z5 * r0[9] + 315.0 * x * z3 * r0[8] +
    315.0 * x * z * r0[7];
  r[2][3][7] =
    x2 * y3 * z7 * r0[12] + 21.0 * x2 * y3 * z5 * r0[11] +
    105.0 * x2 * y3 * z3 * r0[10] + 105.0 * x2 * y3 * z * r0[9] +
    3.0 * x2 * y * z7 * r0[11] + 63.0 * x2 * y * z5 * r0[10] +
    315.0 * x2 * y * z3 * r0[9] + 315.0 * x2 * y * z * r0[8] +
    y3 * z7 * r0[11] + 21.0 * y3 * z5 * r0[10] + 105.0 * y3 * z3 * r0[9] +
    105.0 * y3 * z * r0[8] + 3.0 * y * z7 * r0[10] + 63.0 * y * z5 * r0[9] +
    315.0 * y * z3 * r0[8] + 315.0 * y * z * r0[7];
  r[3][2][7] =
    x3 * y2 * z7 * r0[12] + 21.0 * x3 * y2 * z5 * r0[11] +
    105.0 * x3 * y2 * z3 * r0[10] + 105.0 * x3 * y2 * z * r0[9] +
    x3 * z7 * r0[11] + 21.0 * x3 * z5 * r0[10] + 105.0 * x3 * z3 * r0[9] +
    105.0 * x3 * z * r0[8] + 3.0 * x * y2 * z7 * r0[11] +
    63.0 * x * y2 * z5 * r0[10] + 315.0 * x * y2 * z3 * r0[9] +
    315.0 * x * y2 * z * r0[8] + 3.0 * x * z7 * r0[10] +
    63.0 * x * z5 * r0[9] + 315.0 * x * z3 * r0[8] + 315.0 * x * z * r0[7];
  r[4][1][7] =
    x4 * y * z7 * r0[12] + 21.0 * x4 * y * z5 * r0[11] +
    105.0 * x4 * y * z3 * r0[10] + 105.0 * x4 * y * z * r0[9] +
    6.0 * x2 * y * z7 * r0[11] + 126.0 * x2 * y * z5 * r0[10] +
    630.0 * x2 * y * z3 * r0[9] + 630.0 * x2 * y * z * r0[8] +
    3.0 * y * z7 * r0[10] + 63.0 * y * z5 * r0[9] + 315.0 * y * z3 * r0[8] +
    315.0 * y * z * r0[7];
  r[5][0][7] =
    x5 * z7 * r0[12] + 21.0 * x5 * z5 * r0[11] + 105.0 * x5 * z3 * r0[10] +
    105.0 * x5 * z * r0[9] + 10.0 * x3 * z7 * r0[11] +
    210.0 * x3 * z5 * r0[10] + 1050.0 * x3 * z3 * r0[9] +
    1050.0 * x3 * z * r0[8] + 15.0 * x * z7 * r0[10] +
    315.0 * x * z5 * r0[9] + 1575.0 * x * z3 * r0[8] + 1575.0 * x * z * r0[7];
  r[0][6][6] =
    y6 * z6 * r0[12] + 15.0 * y6 * z4 * r0[11] + 45.0 * y6 * z2 * r0[10] +
    15.0 * y6 * r0[9] + 15.0 * y4 * z6 * r0[11] + 225.0 * y4 * z4 * r0[10] +
    675.0 * y4 * z2 * r0[9] + 225.0 * y4 * r0[8] + 45.0 * y2 * z6 * r0[10] +
    675.0 * y2 * z4 * r0[9] + 2025.0 * y2 * z2 * r0[8] + 675.0 * y2 * r0[7] +
    15.0 * z6 * r0[9] + 225.0 * z4 * r0[8] + 675.0 * z2 * r0[7] +
    225.0 * r0[6];
  r[1][5][6] =
    x * y5 * z6 * r0[12] + 15.0 * x * y5 * z4 * r0[11] +
    45.0 * x * y5 * z2 * r0[10] + 15.0 * x * y5 * r0[9] +
    10.0 * x * y3 * z6 * r0[11] + 150.0 * x * y3 * z4 * r0[10] +
    450.0 * x * y3 * z2 * r0[9] + 150.0 * x * y3 * r0[8] +
    15.0 * x * y * z6 * r0[10] + 225.0 * x * y * z4 * r0[9] +
    675.0 * x * y * z2 * r0[8] + 225.0 * x * y * r0[7];
  r[2][4][6] =
    x2 * y4 * z6 * r0[12] + 15.0 * x2 * y4 * z4 * r0[11] +
    45.0 * x2 * y4 * z2 * r0[10] + 15.0 * x2 * y4 * r0[9] +
    6.0 * x2 * y2 * z6 * r0[11] + 90.0 * x2 * y2 * z4 * r0[10] +
    270.0 * x2 * y2 * z2 * r0[9] + 90.0 * x2 * y2 * r0[8] +
    3.0 * x2 * z6 * r0[10] + 45.0 * x2 * z4 * r0[9] +
    135.0 * x2 * z2 * r0[8] + 45.0 * x2 * r0[7] + y4 * z6 * r0[11] +
    15.0 * y4 * z4 * r0[10] + 45.0 * y4 * z2 * r0[9] + 15.0 * y4 * r0[8] +
    6.0 * y2 * z6 * r0[10] + 90.0 * y2 * z4 * r0[9] +
    270.0 * y2 * z2 * r0[8] + 90.0 * y2 * r0[7] + 3.0 * z6 * r0[9] +
    45.0 * z4 * r0[8] + 135.0 * z2 * r0[7] + 45.0 * r0[6];
  r[3][3][6] =
    x3 * y3 * z6 * r0[12] + 15.0 * x3 * y3 * z4 * r0[11] +
    45.0 * x3 * y3 * z2 * r0[10] + 15.0 * x3 * y3 * r0[9] +
    3.0 * x3 * y * z6 * r0[11] + 45.0 * x3 * y * z4 * r0[10] +
    135.0 * x3 * y * z2 * r0[9] + 45.0 * x3 * y * r0[8] +
    3.0 * x * y3 * z6 * r0[11] + 45.0 * x * y3 * z4 * r0[10] +
    135.0 * x * y3 * z2 * r0[9] + 45.0 * x * y3 * r0[8] +
    9.0 * x * y * z6 * r0[10] + 135.0 * x * y * z4 * r0[9] +
    405.0 * x * y * z2 * r0[8] + 135.0 * x * y * r0[7];
  r[4][2][6] =
    x4 * y2 * z6 * r0[12] + 15.0 * x4 * y2 * z4 * r0[11] +
    45.0 * x4 * y2 * z2 * r0[10] + 15.0 * x4 * y2 * r0[9] + x4 * z6 * r0[11] +
    15.0 * x4 * z4 * r0[10] + 45.0 * x4 * z2 * r0[9] + 15.0 * x4 * r0[8] +
    6.0 * x2 * y2 * z6 * r0[11] + 90.0 * x2 * y2 * z4 * r0[10] +
    270.0 * x2 * y2 * z2 * r0[9] + 90.0 * x2 * y2 * r0[8] +
    6.0 * x2 * z6 * r0[10] + 90.0 * x2 * z4 * r0[9] +
    270.0 * x2 * z2 * r0[8] + 90.0 * x2 * r0[7] + 3.0 * y2 * z6 * r0[10] +
    45.0 * y2 * z4 * r0[9] + 135.0 * y2 * z2 * r0[8] + 45.0 * y2 * r0[7] +
    3.0 * z6 * r0[9] + 45.0 * z4 * r0[8] + 135.0 * z2 * r0[7] + 45.0 * r0[6];
  r[5][1][6] =
    x5 * y * z6 * r0[12] + 15.0 * x5 * y * z4 * r0[11] +
    45.0 * x5 * y * z2 * r0[10] + 15.0 * x5 * y * r0[9] +
    10.0 * x3 * y * z6 * r0[11] + 150.0 * x3 * y * z4 * r0[10] +
    450.0 * x3 * y * z2 * r0[9] + 150.0 * x3 * y * r0[8] +
    15.0 * x * y * z6 * r0[10] + 225.0 * x * y * z4 * r0[9] +
    675.0 * x * y * z2 * r0[8] + 225.0 * x * y * r0[7];
  r[6][0][6] =
    x6 * z6 * r0[12] + 15.0 * x6 * z4 * r0[11] + 45.0 * x6 * z2 * r0[10] +
    15.0 * x6 * r0[9] + 15.0 * x4 * z6 * r0[11] + 225.0 * x4 * z4 * r0[10] +
    675.0 * x4 * z2 * r0[9] + 225.0 * x4 * r0[8] + 45.0 * x2 * z6 * r0[10] +
    675.0 * x2 * z4 * r0[9] + 2025.0 * x2 * z2 * r0[8] + 675.0 * x2 * r0[7] +
    15.0 * z6 * r0[9] + 225.0 * z4 * r0[8] + 675.0 * z2 * r0[7] +
    225.0 * r0[6];
  r[0][7][5] =
    y7 * z5 * r0[12] + 10.0 * y7 * z3 * r0[11] + 15.0 * y7 * z * r0[10] +
    21.0 * y5 * z5 * r0[11] + 210.0 * y5 * z3 * r0[10] +
    315.0 * y5 * z * r0[9] + 105.0 * y3 * z5 * r0[10] +
    1050.0 * y3 * z3 * r0[9] + 1575.0 * y3 * z * r0[8] +
    105.0 * y * z5 * r0[9] + 1050.0 * y * z3 * r0[8] + 1575.0 * y * z * r0[7];
  r[1][6][5] =
    x * y6 * z5 * r0[12] + 10.0 * x * y6 * z3 * r0[11] +
    15.0 * x * y6 * z * r0[10] + 15.0 * x * y4 * z5 * r0[11] +
    150.0 * x * y4 * z3 * r0[10] + 225.0 * x * y4 * z * r0[9] +
    45.0 * x * y2 * z5 * r0[10] + 450.0 * x * y2 * z3 * r0[9] +
    675.0 * x * y2 * z * r0[8] + 15.0 * x * z5 * r0[9] +
    150.0 * x * z3 * r0[8] + 225.0 * x * z * r0[7];
  r[2][5][5] =
    x2 * y5 * z5 * r0[12] + 10.0 * x2 * y5 * z3 * r0[11] +
    15.0 * x2 * y5 * z * r0[10] + 10.0 * x2 * y3 * z5 * r0[11] +
    100.0 * x2 * y3 * z3 * r0[10] + 150.0 * x2 * y3 * z * r0[9] +
    15.0 * x2 * y * z5 * r0[10] + 150.0 * x2 * y * z3 * r0[9] +
    225.0 * x2 * y * z * r0[8] + y5 * z5 * r0[11] + 10.0 * y5 * z3 * r0[10] +
    15.0 * y5 * z * r0[9] + 10.0 * y3 * z5 * r0[10] +
    100.0 * y3 * z3 * r0[9] + 150.0 * y3 * z * r0[8] + 15.0 * y * z5 * r0[9] +
    150.0 * y * z3 * r0[8] + 225.0 * y * z * r0[7];
  r[3][4][5] =
    x3 * y4 * z5 * r0[12] + 10.0 * x3 * y4 * z3 * r0[11] +
    15.0 * x3 * y4 * z * r0[10] + 6.0 * x3 * y2 * z5 * r0[11] +
    60.0 * x3 * y2 * z3 * r0[10] + 90.0 * x3 * y2 * z * r0[9] +
    3.0 * x3 * z5 * r0[10] + 30.0 * x3 * z3 * r0[9] + 45.0 * x3 * z * r0[8] +
    3.0 * x * y4 * z5 * r0[11] + 30.0 * x * y4 * z3 * r0[10] +
    45.0 * x * y4 * z * r0[9] + 18.0 * x * y2 * z5 * r0[10] +
    180.0 * x * y2 * z3 * r0[9] + 270.0 * x * y2 * z * r0[8] +
    9.0 * x * z5 * r0[9] + 90.0 * x * z3 * r0[8] + 135.0 * x * z * r0[7];
  r[4][3][5] =
    x4 * y3 * z5 * r0[12] + 10.0 * x4 * y3 * z3 * r0[11] +
    15.0 * x4 * y3 * z * r0[10] + 3.0 * x4 * y * z5 * r0[11] +
    30.0 * x4 * y * z3 * r0[10] + 45.0 * x4 * y * z * r0[9] +
    6.0 * x2 * y3 * z5 * r0[11] + 60.0 * x2 * y3 * z3 * r0[10] +
    90.0 * x2 * y3 * z * r0[9] + 18.0 * x2 * y * z5 * r0[10] +
    180.0 * x2 * y * z3 * r0[9] + 270.0 * x2 * y * z * r0[8] +
    3.0 * y3 * z5 * r0[10] + 30.0 * y3 * z3 * r0[9] + 45.0 * y3 * z * r0[8] +
    9.0 * y * z5 * r0[9] + 90.0 * y * z3 * r0[8] + 135.0 * y * z * r0[7];
  r[5][2][5] =
    x5 * y2 * z5 * r0[12] + 10.0 * x5 * y2 * z3 * r0[11] +
    15.0 * x5 * y2 * z * r0[10] + x5 * z5 * r0[11] + 10.0 * x5 * z3 * r0[10] +
    15.0 * x5 * z * r0[9] + 10.0 * x3 * y2 * z5 * r0[11] +
    100.0 * x3 * y2 * z3 * r0[10] + 150.0 * x3 * y2 * z * r0[9] +
    10.0 * x3 * z5 * r0[10] + 100.0 * x3 * z3 * r0[9] +
    150.0 * x3 * z * r0[8] + 15.0 * x * y2 * z5 * r0[10] +
    150.0 * x * y2 * z3 * r0[9] + 225.0 * x * y2 * z * r0[8] +
    15.0 * x * z5 * r0[9] + 150.0 * x * z3 * r0[8] + 225.0 * x * z * r0[7];
  r[6][1][5] =
    x6 * y * z5 * r0[12] + 10.0 * x6 * y * z3 * r0[11] +
    15.0 * x6 * y * z * r0[10] + 15.0 * x4 * y * z5 * r0[11] +
    150.0 * x4 * y * z3 * r0[10] + 225.0 * x4 * y * z * r0[9] +
    45.0 * x2 * y * z5 * r0[10] + 450.0 * x2 * y * z3 * r0[9] +
    675.0 * x2 * y * z * r0[8] + 15.0 * y * z5 * r0[9] +
    150.0 * y * z3 * r0[8] + 225.0 * y * z * r0[7];
  r[7][0][5] =
    x7 * z5 * r0[12] + 10.0 * x7 * z3 * r0[11] + 15.0 * x7 * z * r0[10] +
    21.0 * x5 * z5 * r0[11] + 210.0 * x5 * z3 * r0[10] +
    315.0 * x5 * z * r0[9] + 105.0 * x3 * z5 * r0[10] +
    1050.0 * x3 * z3 * r0[9] + 1575.0 * x3 * z * r0[8] +
    105.0 * x * z5 * r0[9] + 1050.0 * x * z3 * r0[8] + 1575.0 * x * z * r0[7];
  r[0][8][4] =
    y8 * z4 * r0[12] + 6.0 * y8 * z2 * r0[11] + 3.0 * y8 * r0[10] +
    28.0 * y6 * z4 * r0[11] + 168.0 * y6 * z2 * r0[10] + 84.0 * y6 * r0[9] +
    210.0 * y4 * z4 * r0[10] + 1260.0 * y4 * z2 * r0[9] + 630.0 * y4 * r0[8] +
    420.0 * y2 * z4 * r0[9] + 2520.0 * y2 * z2 * r0[8] + 1260.0 * y2 * r0[7] +
    105.0 * z4 * r0[8] + 630.0 * z2 * r0[7] + 315.0 * r0[6];
  r[1][7][4] =
    x * y7 * z4 * r0[12] + 6.0 * x * y7 * z2 * r0[11] +
    3.0 * x * y7 * r0[10] + 21.0 * x * y5 * z4 * r0[11] +
    126.0 * x * y5 * z2 * r0[10] + 63.0 * x * y5 * r0[9] +
    105.0 * x * y3 * z4 * r0[10] + 630.0 * x * y3 * z2 * r0[9] +
    315.0 * x * y3 * r0[8] + 105.0 * x * y * z4 * r0[9] +
    630.0 * x * y * z2 * r0[8] + 315.0 * x * y * r0[7];
  r[2][6][4] =
    x2 * y6 * z4 * r0[12] + 6.0 * x2 * y6 * z2 * r0[11] +
    3.0 * x2 * y6 * r0[10] + 15.0 * x2 * y4 * z4 * r0[11] +
    90.0 * x2 * y4 * z2 * r0[10] + 45.0 * x2 * y4 * r0[9] +
    45.0 * x2 * y2 * z4 * r0[10] + 270.0 * x2 * y2 * z2 * r0[9] +
    135.0 * x2 * y2 * r0[8] + 15.0 * x2 * z4 * r0[9] +
    90.0 * x2 * z2 * r0[8] + 45.0 * x2 * r0[7] + y6 * z4 * r0[11] +
    6.0 * y6 * z2 * r0[10] + 3.0 * y6 * r0[9] + 15.0 * y4 * z4 * r0[10] +
    90.0 * y4 * z2 * r0[9] + 45.0 * y4 * r0[8] + 45.0 * y2 * z4 * r0[9] +
    270.0 * y2 * z2 * r0[8] + 135.0 * y2 * r0[7] + 15.0 * z4 * r0[8] +
    90.0 * z2 * r0[7] + 45.0 * r0[6];
  r[3][5][4] =
    x3 * y5 * z4 * r0[12] + 6.0 * x3 * y5 * z2 * r0[11] +
    3.0 * x3 * y5 * r0[10] + 10.0 * x3 * y3 * z4 * r0[11] +
    60.0 * x3 * y3 * z2 * r0[10] + 30.0 * x3 * y3 * r0[9] +
    15.0 * x3 * y * z4 * r0[10] + 90.0 * x3 * y * z2 * r0[9] +
    45.0 * x3 * y * r0[8] + 3.0 * x * y5 * z4 * r0[11] +
    18.0 * x * y5 * z2 * r0[10] + 9.0 * x * y5 * r0[9] +
    30.0 * x * y3 * z4 * r0[10] + 180.0 * x * y3 * z2 * r0[9] +
    90.0 * x * y3 * r0[8] + 45.0 * x * y * z4 * r0[9] +
    270.0 * x * y * z2 * r0[8] + 135.0 * x * y * r0[7];
  r[4][4][4] =
    x4 * y4 * z4 * r0[12] + 6.0 * x4 * y4 * z2 * r0[11] +
    3.0 * x4 * y4 * r0[10] + 6.0 * x4 * y2 * z4 * r0[11] +
    36.0 * x4 * y2 * z2 * r0[10] + 18.0 * x4 * y2 * r0[9] +
    3.0 * x4 * z4 * r0[10] + 18.0 * x4 * z2 * r0[9] + 9.0 * x4 * r0[8] +
    6.0 * x2 * y4 * z4 * r0[11] + 36.0 * x2 * y4 * z2 * r0[10] +
    18.0 * x2 * y4 * r0[9] + 36.0 * x2 * y2 * z4 * r0[10] +
    216.0 * x2 * y2 * z2 * r0[9] + 108.0 * x2 * y2 * r0[8] +
    18.0 * x2 * z4 * r0[9] + 108.0 * x2 * z2 * r0[8] + 54.0 * x2 * r0[7] +
    3.0 * y4 * z4 * r0[10] + 18.0 * y4 * z2 * r0[9] + 9.0 * y4 * r0[8] +
    18.0 * y2 * z4 * r0[9] + 108.0 * y2 * z2 * r0[8] + 54.0 * y2 * r0[7] +
    9.0 * z4 * r0[8] + 54.0 * z2 * r0[7] + 27.0 * r0[6];
  r[5][3][4] =
    x5 * y3 * z4 * r0[12] + 6.0 * x5 * y3 * z2 * r0[11] +
    3.0 * x5 * y3 * r0[10] + 3.0 * x5 * y * z4 * r0[11] +
    18.0 * x5 * y * z2 * r0[10] + 9.0 * x5 * y * r0[9] +
    10.0 * x3 * y3 * z4 * r0[11] + 60.0 * x3 * y3 * z2 * r0[10] +
    30.0 * x3 * y3 * r0[9] + 30.0 * x3 * y * z4 * r0[10] +
    180.0 * x3 * y * z2 * r0[9] + 90.0 * x3 * y * r0[8] +
    15.0 * x * y3 * z4 * r0[10] + 90.0 * x * y3 * z2 * r0[9] +
    45.0 * x * y3 * r0[8] + 45.0 * x * y * z4 * r0[9] +
    270.0 * x * y * z2 * r0[8] + 135.0 * x * y * r0[7];
  r[6][2][4] =
    x6 * y2 * z4 * r0[12] + 6.0 * x6 * y2 * z2 * r0[11] +
    3.0 * x6 * y2 * r0[10] + x6 * z4 * r0[11] + 6.0 * x6 * z2 * r0[10] +
    3.0 * x6 * r0[9] + 15.0 * x4 * y2 * z4 * r0[11] +
    90.0 * x4 * y2 * z2 * r0[10] + 45.0 * x4 * y2 * r0[9] +
    15.0 * x4 * z4 * r0[10] + 90.0 * x4 * z2 * r0[9] + 45.0 * x4 * r0[8] +
    45.0 * x2 * y2 * z4 * r0[10] + 270.0 * x2 * y2 * z2 * r0[9] +
    135.0 * x2 * y2 * r0[8] + 45.0 * x2 * z4 * r0[9] +
    270.0 * x2 * z2 * r0[8] + 135.0 * x2 * r0[7] + 15.0 * y2 * z4 * r0[9] +
    90.0 * y2 * z2 * r0[8] + 45.0 * y2 * r0[7] + 15.0 * z4 * r0[8] +
    90.0 * z2 * r0[7] + 45.0 * r0[6];
  r[7][1][4] =
    x7 * y * z4 * r0[12] + 6.0 * x7 * y * z2 * r0[11] +
    3.0 * x7 * y * r0[10] + 21.0 * x5 * y * z4 * r0[11] +
    126.0 * x5 * y * z2 * r0[10] + 63.0 * x5 * y * r0[9] +
    105.0 * x3 * y * z4 * r0[10] + 630.0 * x3 * y * z2 * r0[9] +
    315.0 * x3 * y * r0[8] + 105.0 * x * y * z4 * r0[9] +
    630.0 * x * y * z2 * r0[8] + 315.0 * x * y * r0[7];
  r[8][0][4] =
    x8 * z4 * r0[12] + 6.0 * x8 * z2 * r0[11] + 3.0 * x8 * r0[10] +
    28.0 * x6 * z4 * r0[11] + 168.0 * x6 * z2 * r0[10] + 84.0 * x6 * r0[9] +
    210.0 * x4 * z4 * r0[10] + 1260.0 * x4 * z2 * r0[9] + 630.0 * x4 * r0[8] +
    420.0 * x2 * z4 * r0[9] + 2520.0 * x2 * z2 * r0[8] + 1260.0 * x2 * r0[7] +
    105.0 * z4 * r0[8] + 630.0 * z2 * r0[7] + 315.0 * r0[6];
  r[0][9][3] =
    y9 * z3 * r0[12] + 3.0 * y9 * z * r0[11] + 36.0 * y7 * z3 * r0[11] +
    108.0 * y7 * z * r0[10] + 378.0 * y5 * z3 * r0[10] +
    1134.0 * y5 * z * r0[9] + 1260.0 * y3 * z3 * r0[9] +
    3780.0 * y3 * z * r0[8] + 945.0 * y * z3 * r0[8] + 2835.0 * y * z * r0[7];
  r[1][8][3] =
    x * y8 * z3 * r0[12] + 3.0 * x * y8 * z * r0[11] +
    28.0 * x * y6 * z3 * r0[11] + 84.0 * x * y6 * z * r0[10] +
    210.0 * x * y4 * z3 * r0[10] + 630.0 * x * y4 * z * r0[9] +
    420.0 * x * y2 * z3 * r0[9] + 1260.0 * x * y2 * z * r0[8] +
    105.0 * x * z3 * r0[8] + 315.0 * x * z * r0[7];
  r[2][7][3] =
    x2 * y7 * z3 * r0[12] + 3.0 * x2 * y7 * z * r0[11] +
    21.0 * x2 * y5 * z3 * r0[11] + 63.0 * x2 * y5 * z * r0[10] +
    105.0 * x2 * y3 * z3 * r0[10] + 315.0 * x2 * y3 * z * r0[9] +
    105.0 * x2 * y * z3 * r0[9] + 315.0 * x2 * y * z * r0[8] +
    y7 * z3 * r0[11] + 3.0 * y7 * z * r0[10] + 21.0 * y5 * z3 * r0[10] +
    63.0 * y5 * z * r0[9] + 105.0 * y3 * z3 * r0[9] + 315.0 * y3 * z * r0[8] +
    105.0 * y * z3 * r0[8] + 315.0 * y * z * r0[7];
  r[3][6][3] =
    x3 * y6 * z3 * r0[12] + 3.0 * x3 * y6 * z * r0[11] +
    15.0 * x3 * y4 * z3 * r0[11] + 45.0 * x3 * y4 * z * r0[10] +
    45.0 * x3 * y2 * z3 * r0[10] + 135.0 * x3 * y2 * z * r0[9] +
    15.0 * x3 * z3 * r0[9] + 45.0 * x3 * z * r0[8] +
    3.0 * x * y6 * z3 * r0[11] + 9.0 * x * y6 * z * r0[10] +
    45.0 * x * y4 * z3 * r0[10] + 135.0 * x * y4 * z * r0[9] +
    135.0 * x * y2 * z3 * r0[9] + 405.0 * x * y2 * z * r0[8] +
    45.0 * x * z3 * r0[8] + 135.0 * x * z * r0[7];
  r[4][5][3] =
    x4 * y5 * z3 * r0[12] + 3.0 * x4 * y5 * z * r0[11] +
    10.0 * x4 * y3 * z3 * r0[11] + 30.0 * x4 * y3 * z * r0[10] +
    15.0 * x4 * y * z3 * r0[10] + 45.0 * x4 * y * z * r0[9] +
    6.0 * x2 * y5 * z3 * r0[11] + 18.0 * x2 * y5 * z * r0[10] +
    60.0 * x2 * y3 * z3 * r0[10] + 180.0 * x2 * y3 * z * r0[9] +
    90.0 * x2 * y * z3 * r0[9] + 270.0 * x2 * y * z * r0[8] +
    3.0 * y5 * z3 * r0[10] + 9.0 * y5 * z * r0[9] + 30.0 * y3 * z3 * r0[9] +
    90.0 * y3 * z * r0[8] + 45.0 * y * z3 * r0[8] + 135.0 * y * z * r0[7];
  r[5][4][3] =
    x5 * y4 * z3 * r0[12] + 3.0 * x5 * y4 * z * r0[11] +
    6.0 * x5 * y2 * z3 * r0[11] + 18.0 * x5 * y2 * z * r0[10] +
    3.0 * x5 * z3 * r0[10] + 9.0 * x5 * z * r0[9] +
    10.0 * x3 * y4 * z3 * r0[11] + 30.0 * x3 * y4 * z * r0[10] +
    60.0 * x3 * y2 * z3 * r0[10] + 180.0 * x3 * y2 * z * r0[9] +
    30.0 * x3 * z3 * r0[9] + 90.0 * x3 * z * r0[8] +
    15.0 * x * y4 * z3 * r0[10] + 45.0 * x * y4 * z * r0[9] +
    90.0 * x * y2 * z3 * r0[9] + 270.0 * x * y2 * z * r0[8] +
    45.0 * x * z3 * r0[8] + 135.0 * x * z * r0[7];
  r[6][3][3] =
    x6 * y3 * z3 * r0[12] + 3.0 * x6 * y3 * z * r0[11] +
    3.0 * x6 * y * z3 * r0[11] + 9.0 * x6 * y * z * r0[10] +
    15.0 * x4 * y3 * z3 * r0[11] + 45.0 * x4 * y3 * z * r0[10] +
    45.0 * x4 * y * z3 * r0[10] + 135.0 * x4 * y * z * r0[9] +
    45.0 * x2 * y3 * z3 * r0[10] + 135.0 * x2 * y3 * z * r0[9] +
    135.0 * x2 * y * z3 * r0[9] + 405.0 * x2 * y * z * r0[8] +
    15.0 * y3 * z3 * r0[9] + 45.0 * y3 * z * r0[8] + 45.0 * y * z3 * r0[8] +
    135.0 * y * z * r0[7];
  r[7][2][3] =
    x7 * y2 * z3 * r0[12] + 3.0 * x7 * y2 * z * r0[11] + x7 * z3 * r0[11] +
    3.0 * x7 * z * r0[10] + 21.0 * x5 * y2 * z3 * r0[11] +
    63.0 * x5 * y2 * z * r0[10] + 21.0 * x5 * z3 * r0[10] +
    63.0 * x5 * z * r0[9] + 105.0 * x3 * y2 * z3 * r0[10] +
    315.0 * x3 * y2 * z * r0[9] + 105.0 * x3 * z3 * r0[9] +
    315.0 * x3 * z * r0[8] + 105.0 * x * y2 * z3 * r0[9] +
    315.0 * x * y2 * z * r0[8] + 105.0 * x * z3 * r0[8] +
    315.0 * x * z * r0[7];
  r[8][1][3] =
    x8 * y * z3 * r0[12] + 3.0 * x8 * y * z * r0[11] +
    28.0 * x6 * y * z3 * r0[11] + 84.0 * x6 * y * z * r0[10] +
    210.0 * x4 * y * z3 * r0[10] + 630.0 * x4 * y * z * r0[9] +
    420.0 * x2 * y * z3 * r0[9] + 1260.0 * x2 * y * z * r0[8] +
    105.0 * y * z3 * r0[8] + 315.0 * y * z * r0[7];
  r[9][0][3] =
    x9 * z3 * r0[12] + 3.0 * x9 * z * r0[11] + 36.0 * x7 * z3 * r0[11] +
    108.0 * x7 * z * r0[10] + 378.0 * x5 * z3 * r0[10] +
    1134.0 * x5 * z * r0[9] + 1260.0 * x3 * z3 * r0[9] +
    3780.0 * x3 * z * r0[8] + 945.0 * x * z3 * r0[8] + 2835.0 * x * z * r0[7];
  r[0][10][2] =
    y10 * z2 * r0[12] + y10 * r0[11] + 45.0 * y8 * z2 * r0[11] +
    45.0 * y8 * r0[10] + 630.0 * y6 * z2 * r0[10] + 630.0 * y6 * r0[9] +
    3150.0 * y4 * z2 * r0[9] + 3150.0 * y4 * r0[8] +
    4725.0 * y2 * z2 * r0[8] + 4725.0 * y2 * r0[7] + 945.0 * z2 * r0[7] +
    945.0 * r0[6];
  r[1][9][2] =
    x * y9 * z2 * r0[12] + x * y9 * r0[11] + 36.0 * x * y7 * z2 * r0[11] +
    36.0 * x * y7 * r0[10] + 378.0 * x * y5 * z2 * r0[10] +
    378.0 * x * y5 * r0[9] + 1260.0 * x * y3 * z2 * r0[9] +
    1260.0 * x * y3 * r0[8] + 945.0 * x * y * z2 * r0[8] +
    945.0 * x * y * r0[7];
  r[2][8][2] =
    x2 * y8 * z2 * r0[12] + x2 * y8 * r0[11] + 28.0 * x2 * y6 * z2 * r0[11] +
    28.0 * x2 * y6 * r0[10] + 210.0 * x2 * y4 * z2 * r0[10] +
    210.0 * x2 * y4 * r0[9] + 420.0 * x2 * y2 * z2 * r0[9] +
    420.0 * x2 * y2 * r0[8] + 105.0 * x2 * z2 * r0[8] + 105.0 * x2 * r0[7] +
    y8 * z2 * r0[11] + y8 * r0[10] + 28.0 * y6 * z2 * r0[10] +
    28.0 * y6 * r0[9] + 210.0 * y4 * z2 * r0[9] + 210.0 * y4 * r0[8] +
    420.0 * y2 * z2 * r0[8] + 420.0 * y2 * r0[7] + 105.0 * z2 * r0[7] +
    105.0 * r0[6];
  r[3][7][2] =
    x3 * y7 * z2 * r0[12] + x3 * y7 * r0[11] + 21.0 * x3 * y5 * z2 * r0[11] +
    21.0 * x3 * y5 * r0[10] + 105.0 * x3 * y3 * z2 * r0[10] +
    105.0 * x3 * y3 * r0[9] + 105.0 * x3 * y * z2 * r0[9] +
    105.0 * x3 * y * r0[8] + 3.0 * x * y7 * z2 * r0[11] +
    3.0 * x * y7 * r0[10] + 63.0 * x * y5 * z2 * r0[10] +
    63.0 * x * y5 * r0[9] + 315.0 * x * y3 * z2 * r0[9] +
    315.0 * x * y3 * r0[8] + 315.0 * x * y * z2 * r0[8] +
    315.0 * x * y * r0[7];
  r[4][6][2] =
    x4 * y6 * z2 * r0[12] + x4 * y6 * r0[11] + 15.0 * x4 * y4 * z2 * r0[11] +
    15.0 * x4 * y4 * r0[10] + 45.0 * x4 * y2 * z2 * r0[10] +
    45.0 * x4 * y2 * r0[9] + 15.0 * x4 * z2 * r0[9] + 15.0 * x4 * r0[8] +
    6.0 * x2 * y6 * z2 * r0[11] + 6.0 * x2 * y6 * r0[10] +
    90.0 * x2 * y4 * z2 * r0[10] + 90.0 * x2 * y4 * r0[9] +
    270.0 * x2 * y2 * z2 * r0[9] + 270.0 * x2 * y2 * r0[8] +
    90.0 * x2 * z2 * r0[8] + 90.0 * x2 * r0[7] + 3.0 * y6 * z2 * r0[10] +
    3.0 * y6 * r0[9] + 45.0 * y4 * z2 * r0[9] + 45.0 * y4 * r0[8] +
    135.0 * y2 * z2 * r0[8] + 135.0 * y2 * r0[7] + 45.0 * z2 * r0[7] +
    45.0 * r0[6];
  r[5][5][2] =
    x5 * y5 * z2 * r0[12] + x5 * y5 * r0[11] + 10.0 * x5 * y3 * z2 * r0[11] +
    10.0 * x5 * y3 * r0[10] + 15.0 * x5 * y * z2 * r0[10] +
    15.0 * x5 * y * r0[9] + 10.0 * x3 * y5 * z2 * r0[11] +
    10.0 * x3 * y5 * r0[10] + 100.0 * x3 * y3 * z2 * r0[10] +
    100.0 * x3 * y3 * r0[9] + 150.0 * x3 * y * z2 * r0[9] +
    150.0 * x3 * y * r0[8] + 15.0 * x * y5 * z2 * r0[10] +
    15.0 * x * y5 * r0[9] + 150.0 * x * y3 * z2 * r0[9] +
    150.0 * x * y3 * r0[8] + 225.0 * x * y * z2 * r0[8] +
    225.0 * x * y * r0[7];
  r[6][4][2] =
    x6 * y4 * z2 * r0[12] + x6 * y4 * r0[11] + 6.0 * x6 * y2 * z2 * r0[11] +
    6.0 * x6 * y2 * r0[10] + 3.0 * x6 * z2 * r0[10] + 3.0 * x6 * r0[9] +
    15.0 * x4 * y4 * z2 * r0[11] + 15.0 * x4 * y4 * r0[10] +
    90.0 * x4 * y2 * z2 * r0[10] + 90.0 * x4 * y2 * r0[9] +
    45.0 * x4 * z2 * r0[9] + 45.0 * x4 * r0[8] +
    45.0 * x2 * y4 * z2 * r0[10] + 45.0 * x2 * y4 * r0[9] +
    270.0 * x2 * y2 * z2 * r0[9] + 270.0 * x2 * y2 * r0[8] +
    135.0 * x2 * z2 * r0[8] + 135.0 * x2 * r0[7] + 15.0 * y4 * z2 * r0[9] +
    15.0 * y4 * r0[8] + 90.0 * y2 * z2 * r0[8] + 90.0 * y2 * r0[7] +
    45.0 * z2 * r0[7] + 45.0 * r0[6];
  r[7][3][2] =
    x7 * y3 * z2 * r0[12] + x7 * y3 * r0[11] + 3.0 * x7 * y * z2 * r0[11] +
    3.0 * x7 * y * r0[10] + 21.0 * x5 * y3 * z2 * r0[11] +
    21.0 * x5 * y3 * r0[10] + 63.0 * x5 * y * z2 * r0[10] +
    63.0 * x5 * y * r0[9] + 105.0 * x3 * y3 * z2 * r0[10] +
    105.0 * x3 * y3 * r0[9] + 315.0 * x3 * y * z2 * r0[9] +
    315.0 * x3 * y * r0[8] + 105.0 * x * y3 * z2 * r0[9] +
    105.0 * x * y3 * r0[8] + 315.0 * x * y * z2 * r0[8] +
    315.0 * x * y * r0[7];
  r[8][2][2] =
    x8 * y2 * z2 * r0[12] + x8 * y2 * r0[11] + x8 * z2 * r0[11] +
    x8 * r0[10] + 28.0 * x6 * y2 * z2 * r0[11] + 28.0 * x6 * y2 * r0[10] +
    28.0 * x6 * z2 * r0[10] + 28.0 * x6 * r0[9] +
    210.0 * x4 * y2 * z2 * r0[10] + 210.0 * x4 * y2 * r0[9] +
    210.0 * x4 * z2 * r0[9] + 210.0 * x4 * r0[8] +
    420.0 * x2 * y2 * z2 * r0[9] + 420.0 * x2 * y2 * r0[8] +
    420.0 * x2 * z2 * r0[8] + 420.0 * x2 * r0[7] + 105.0 * y2 * z2 * r0[8] +
    105.0 * y2 * r0[7] + 105.0 * z2 * r0[7] + 105.0 * r0[6];
  r[9][1][2] =
    x9 * y * z2 * r0[12] + x9 * y * r0[11] + 36.0 * x7 * y * z2 * r0[11] +
    36.0 * x7 * y * r0[10] + 378.0 * x5 * y * z2 * r0[10] +
    378.0 * x5 * y * r0[9] + 1260.0 * x3 * y * z2 * r0[9] +
    1260.0 * x3 * y * r0[8] + 945.0 * x * y * z2 * r0[8] +
    945.0 * x * y * r0[7];
  r[10][0][2] =
    x10 * z2 * r0[12] + x10 * r0[11] + 45.0 * x8 * z2 * r0[11] +
    45.0 * x8 * r0[10] + 630.0 * x6 * z2 * r0[10] + 630.0 * x6 * r0[9] +
    3150.0 * x4 * z2 * r0[9] + 3150.0 * x4 * r0[8] +
    4725.0 * x2 * z2 * r0[8] + 4725.0 * x2 * r0[7] + 945.0 * z2 * r0[7] +
    945.0 * r0[6];
  r[0][11][1] =
    y11 * z * r0[12] + 55.0 * y9 * z * r0[11] + 990.0 * y7 * z * r0[10] +
    6930.0 * y5 * z * r0[9] + 17325.0 * y3 * z * r0[8] +
    10395.0 * y * z * r0[7];
  r[1][10][1] =
    x * y10 * z * r0[12] + 45.0 * x * y8 * z * r0[11] +
    630.0 * x * y6 * z * r0[10] + 3150.0 * x * y4 * z * r0[9] +
    4725.0 * x * y2 * z * r0[8] + 945.0 * x * z * r0[7];
  r[2][9][1] =
    x2 * y9 * z * r0[12] + 36.0 * x2 * y7 * z * r0[11] +
    378.0 * x2 * y5 * z * r0[10] + 1260.0 * x2 * y3 * z * r0[9] +
    945.0 * x2 * y * z * r0[8] + y9 * z * r0[11] + 36.0 * y7 * z * r0[10] +
    378.0 * y5 * z * r0[9] + 1260.0 * y3 * z * r0[8] + 945.0 * y * z * r0[7];
  r[3][8][1] =
    x3 * y8 * z * r0[12] + 28.0 * x3 * y6 * z * r0[11] +
    210.0 * x3 * y4 * z * r0[10] + 420.0 * x3 * y2 * z * r0[9] +
    105.0 * x3 * z * r0[8] + 3.0 * x * y8 * z * r0[11] +
    84.0 * x * y6 * z * r0[10] + 630.0 * x * y4 * z * r0[9] +
    1260.0 * x * y2 * z * r0[8] + 315.0 * x * z * r0[7];
  r[4][7][1] =
    x4 * y7 * z * r0[12] + 21.0 * x4 * y5 * z * r0[11] +
    105.0 * x4 * y3 * z * r0[10] + 105.0 * x4 * y * z * r0[9] +
    6.0 * x2 * y7 * z * r0[11] + 126.0 * x2 * y5 * z * r0[10] +
    630.0 * x2 * y3 * z * r0[9] + 630.0 * x2 * y * z * r0[8] +
    3.0 * y7 * z * r0[10] + 63.0 * y5 * z * r0[9] + 315.0 * y3 * z * r0[8] +
    315.0 * y * z * r0[7];
  r[5][6][1] =
    x5 * y6 * z * r0[12] + 15.0 * x5 * y4 * z * r0[11] +
    45.0 * x5 * y2 * z * r0[10] + 15.0 * x5 * z * r0[9] +
    10.0 * x3 * y6 * z * r0[11] + 150.0 * x3 * y4 * z * r0[10] +
    450.0 * x3 * y2 * z * r0[9] + 150.0 * x3 * z * r0[8] +
    15.0 * x * y6 * z * r0[10] + 225.0 * x * y4 * z * r0[9] +
    675.0 * x * y2 * z * r0[8] + 225.0 * x * z * r0[7];
  r[6][5][1] =
    x6 * y5 * z * r0[12] + 10.0 * x6 * y3 * z * r0[11] +
    15.0 * x6 * y * z * r0[10] + 15.0 * x4 * y5 * z * r0[11] +
    150.0 * x4 * y3 * z * r0[10] + 225.0 * x4 * y * z * r0[9] +
    45.0 * x2 * y5 * z * r0[10] + 450.0 * x2 * y3 * z * r0[9] +
    675.0 * x2 * y * z * r0[8] + 15.0 * y5 * z * r0[9] +
    150.0 * y3 * z * r0[8] + 225.0 * y * z * r0[7];
  r[7][4][1] =
    x7 * y4 * z * r0[12] + 6.0 * x7 * y2 * z * r0[11] +
    3.0 * x7 * z * r0[10] + 21.0 * x5 * y4 * z * r0[11] +
    126.0 * x5 * y2 * z * r0[10] + 63.0 * x5 * z * r0[9] +
    105.0 * x3 * y4 * z * r0[10] + 630.0 * x3 * y2 * z * r0[9] +
    315.0 * x3 * z * r0[8] + 105.0 * x * y4 * z * r0[9] +
    630.0 * x * y2 * z * r0[8] + 315.0 * x * z * r0[7];
  r[8][3][1] =
    x8 * y3 * z * r0[12] + 3.0 * x8 * y * z * r0[11] +
    28.0 * x6 * y3 * z * r0[11] + 84.0 * x6 * y * z * r0[10] +
    210.0 * x4 * y3 * z * r0[10] + 630.0 * x4 * y * z * r0[9] +
    420.0 * x2 * y3 * z * r0[9] + 1260.0 * x2 * y * z * r0[8] +
    105.0 * y3 * z * r0[8] + 315.0 * y * z * r0[7];
  r[9][2][1] =
    x9 * y2 * z * r0[12] + x9 * z * r0[11] + 36.0 * x7 * y2 * z * r0[11] +
    36.0 * x7 * z * r0[10] + 378.0 * x5 * y2 * z * r0[10] +
    378.0 * x5 * z * r0[9] + 1260.0 * x3 * y2 * z * r0[9] +
    1260.0 * x3 * z * r0[8] + 945.0 * x * y2 * z * r0[8] +
    945.0 * x * z * r0[7];
  r[10][1][1] =
    x10 * y * z * r0[12] + 45.0 * x8 * y * z * r0[11] +
    630.0 * x6 * y * z * r0[10] + 3150.0 * x4 * y * z * r0[9] +
    4725.0 * x2 * y * z * r0[8] + 945.0 * y * z * r0[7];
  r[11][0][1] =
    x11 * z * r0[12] + 55.0 * x9 * z * r0[11] + 990.0 * x7 * z * r0[10] +
    6930.0 * x5 * z * r0[9] + 17325.0 * x3 * z * r0[8] +
    10395.0 * x * z * r0[7];
  r[0][12][0] =
    y12 * r0[12] + 66.0 * y10 * r0[11] + 1485.0 * y8 * r0[10] +
    13860.0 * y6 * r0[9] + 51975.0 * y4 * r0[8] + 62370.0 * y2 * r0[7] +
    10395.0 * r0[6];
  r[1][11][0] =
    x * y11 * r0[12] + 55.0 * x * y9 * r0[11] + 990.0 * x * y7 * r0[10] +
    6930.0 * x * y5 * r0[9] + 17325.0 * x * y3 * r0[8] +
    10395.0 * x * y * r0[7];
  r[2][10][0] =
    x2 * y10 * r0[12] + 45.0 * x2 * y8 * r0[11] + 630.0 * x2 * y6 * r0[10] +
    3150.0 * x2 * y4 * r0[9] + 4725.0 * x2 * y2 * r0[8] + 945.0 * x2 * r0[7] +
    y10 * r0[11] + 45.0 * y8 * r0[10] + 630.0 * y6 * r0[9] +
    3150.0 * y4 * r0[8] + 4725.0 * y2 * r0[7] + 945.0 * r0[6];
  r[3][9][0] =
    x3 * y9 * r0[12] + 36.0 * x3 * y7 * r0[11] + 378.0 * x3 * y5 * r0[10] +
    1260.0 * x3 * y3 * r0[9] + 945.0 * x3 * y * r0[8] +
    3.0 * x * y9 * r0[11] + 108.0 * x * y7 * r0[10] +
    1134.0 * x * y5 * r0[9] + 3780.0 * x * y3 * r0[8] +
    2835.0 * x * y * r0[7];
  r[4][8][0] =
    x4 * y8 * r0[12] + 28.0 * x4 * y6 * r0[11] + 210.0 * x4 * y4 * r0[10] +
    420.0 * x4 * y2 * r0[9] + 105.0 * x4 * r0[8] + 6.0 * x2 * y8 * r0[11] +
    168.0 * x2 * y6 * r0[10] + 1260.0 * x2 * y4 * r0[9] +
    2520.0 * x2 * y2 * r0[8] + 630.0 * x2 * r0[7] + 3.0 * y8 * r0[10] +
    84.0 * y6 * r0[9] + 630.0 * y4 * r0[8] + 1260.0 * y2 * r0[7] +
    315.0 * r0[6];
  r[5][7][0] =
    x5 * y7 * r0[12] + 21.0 * x5 * y5 * r0[11] + 105.0 * x5 * y3 * r0[10] +
    105.0 * x5 * y * r0[9] + 10.0 * x3 * y7 * r0[11] +
    210.0 * x3 * y5 * r0[10] + 1050.0 * x3 * y3 * r0[9] +
    1050.0 * x3 * y * r0[8] + 15.0 * x * y7 * r0[10] +
    315.0 * x * y5 * r0[9] + 1575.0 * x * y3 * r0[8] + 1575.0 * x * y * r0[7];
  r[6][6][0] =
    x6 * y6 * r0[12] + 15.0 * x6 * y4 * r0[11] + 45.0 * x6 * y2 * r0[10] +
    15.0 * x6 * r0[9] + 15.0 * x4 * y6 * r0[11] + 225.0 * x4 * y4 * r0[10] +
    675.0 * x4 * y2 * r0[9] + 225.0 * x4 * r0[8] + 45.0 * x2 * y6 * r0[10] +
    675.0 * x2 * y4 * r0[9] + 2025.0 * x2 * y2 * r0[8] + 675.0 * x2 * r0[7] +
    15.0 * y6 * r0[9] + 225.0 * y4 * r0[8] + 675.0 * y2 * r0[7] +
    225.0 * r0[6];
  r[7][5][0] =
    x7 * y5 * r0[12] + 10.0 * x7 * y3 * r0[11] + 15.0 * x7 * y * r0[10] +
    21.0 * x5 * y5 * r0[11] + 210.0 * x5 * y3 * r0[10] +
    315.0 * x5 * y * r0[9] + 105.0 * x3 * y5 * r0[10] +
    1050.0 * x3 * y3 * r0[9] + 1575.0 * x3 * y * r0[8] +
    105.0 * x * y5 * r0[9] + 1050.0 * x * y3 * r0[8] + 1575.0 * x * y * r0[7];
  r[8][4][0] =
    x8 * y4 * r0[12] + 6.0 * x8 * y2 * r0[11] + 3.0 * x8 * r0[10] +
    28.0 * x6 * y4 * r0[11] + 168.0 * x6 * y2 * r0[10] + 84.0 * x6 * r0[9] +
    210.0 * x4 * y4 * r0[10] + 1260.0 * x4 * y2 * r0[9] + 630.0 * x4 * r0[8] +
    420.0 * x2 * y4 * r0[9] + 2520.0 * x2 * y2 * r0[8] + 1260.0 * x2 * r0[7] +
    105.0 * y4 * r0[8] + 630.0 * y2 * r0[7] + 315.0 * r0[6];
  r[9][3][0] =
    x9 * y3 * r0[12] + 3.0 * x9 * y * r0[11] + 36.0 * x7 * y3 * r0[11] +
    108.0 * x7 * y * r0[10] + 378.0 * x5 * y3 * r0[10] +
    1134.0 * x5 * y * r0[9] + 1260.0 * x3 * y3 * r0[9] +
    3780.0 * x3 * y * r0[8] + 945.0 * x * y3 * r0[8] + 2835.0 * x * y * r0[7];
  r[10][2][0] =
    x10 * y2 * r0[12] + x10 * r0[11] + 45.0 * x8 * y2 * r0[11] +
    45.0 * x8 * r0[10] + 630.0 * x6 * y2 * r0[10] + 630.0 * x6 * r0[9] +
    3150.0 * x4 * y2 * r0[9] + 3150.0 * x4 * r0[8] +
    4725.0 * x2 * y2 * r0[8] + 4725.0 * x2 * r0[7] + 945.0 * y2 * r0[7] +
    945.0 * r0[6];
  r[11][1][0] =
    x11 * y * r0[12] + 55.0 * x9 * y * r0[11] + 990.0 * x7 * y * r0[10] +
    6930.0 * x5 * y * r0[9] + 17325.0 * x3 * y * r0[8] +
    10395.0 * x * y * r0[7];
  r[12][0][0] =
    x12 * r0[12] + 66.0 * x10 * r0[11] + 1485.0 * x8 * r0[10] +
    13860.0 * x6 * r0[9] + 51975.0 * x4 * r0[8] + 62370.0 * x2 * r0[7] +
    10395.0 * r0[6];
  if (ltot == 12)
    {
      return;
    }
  fatalError ("MAXL exceeded in formMDR\n");
}
#endif
